from PyQt5.QtCore import QRectF, pyqtSignal, Qt, QPoint
from PyQt5.QtGui import QMouseEvent, QKeyEvent
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QGraphicsView

from urh import constants
from urh.SceneManager import SceneManager
from urh.ui.ROI import ROI
from urh.ui.ZoomableScene import ZoomableScene


class SelectableGraphicView(QGraphicsView):
    sep_area_moving = pyqtSignal(float)
    sep_area_changed = pyqtSignal(float)
    sel_area_width_changed = pyqtSignal(int)
    sel_area_start_end_changed = pyqtSignal(int, int)
    shift_state_changed = pyqtSignal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setResizeAnchor(QGraphicsView.NoAnchor)
        self.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.setRenderHints(QPainter.Antialiasing)

        self.scene_manager = None  # type: SceneManager
        self.mouse_press_pos = None  # type: QPoint
        self.mouse_pos = None  # type: QPoint
        self.grab_start = None  # type: QPoint

        self.xmove = 0

        self.separation_area_moving = False

        self.shift_mode = False  # Shift Key currently pressed?

    def scene(self) -> ZoomableScene:
        return super().scene()

    @property
    def selection_area(self) -> ROI:
        return self.scene().selection_area

    @selection_area.setter
    def selection_area(self, value):
        self.scene().selection_area = value

    @property
    def hold_shift_to_drag(self) -> bool:
        return constants.SETTINGS.value('hold_shift_to_drag', False, type=bool)

    def is_pos_in_separea(self, pos: QPoint):
        """
        GraphicViews can override this, if they need a seperation area.
        E.g. EpicGraphic View will do for Demodulated View

        :param pos:
        :return:
        """

        return False

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Shift:
            self.shift_mode = True
            self.shift_state_changed.emit(True)

            if self.hold_shift_to_drag:
                self.setCursor(Qt.OpenHandCursor)
            else:
                self.unsetCursor()
                self.grab_start = None

        super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Shift:
            self.shift_mode = False
            self.shift_state_changed.emit(False)

            if self.hold_shift_to_drag:
                self.unsetCursor()
                self.grab_start = None
            else:
                self.setCursor(Qt.OpenHandCursor)

        super().keyPressEvent(event)

    def mousePressEvent(self, event: QMouseEvent):
        if self.scene() is None:
            return

        cursor = self.cursor().shape()
        has_shift_modifier = event.modifiers() == Qt.ShiftModifier
        is_in_shift_mode = (has_shift_modifier and self.hold_shift_to_drag) \
                           or (not has_shift_modifier and not self.hold_shift_to_drag) \
                              and cursor != Qt.SplitHCursor and cursor != Qt.SplitVCursor

        if event.buttons() == Qt.LeftButton and is_in_shift_mode:
            self.setCursor(Qt.ClosedHandCursor)
            self.grab_start = event.pos()
        elif event.buttons() == Qt.LeftButton:
            if self.is_pos_in_separea(self.mapToScene(event.pos())):
                self.separation_area_moving = True
                self.setCursor(Qt.SplitVCursor)

            elif self.selection_area.is_empty or self.selection_area.selected_edge is None:
                # Neue ROI anlegen
                self.mouse_press_pos = event.pos()
                self.mouse_pos = event.pos()
                scene_pos = self.mapToScene(self.mouse_press_pos)
                x = scene_pos.x()
                self.set_selection_area(x=x, w=0)
                self.selection_area.finished = False

            elif self.selection_area.selected_edge is not None:
                self.selection_area.resizing = True

    def mouseMoveEvent(self, event: QMouseEvent):
        if self.scene() is None:
            return

        cursor = self.cursor().shape()

        if self.grab_start is not None:
            move = self.grab_start.x() - event.pos().x()
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + move)
            self.grab_start = event.pos()
            return

        if self.separation_area_moving:
            y_sep = self.mapToScene(event.pos()).y()
            y = self.sceneRect().y()
            h = self.sceneRect().height()
            if y < y_sep < y + h:
                self.scene().draw_sep_area(y_sep)
                self.sep_area_moving.emit(y_sep)
        elif self.is_pos_in_separea(self.mapToScene(event.pos())):
            self.setCursor(Qt.SplitVCursor)
        elif cursor == Qt.SplitVCursor:
            self.unsetCursor()

        if self.selection_area.finished and not self.selection_area.resizing:
            pos = self.mapToScene(event.pos())
            roi_edge = self.selection_area.get_selected_edge(pos, self.view_rect().width())
            if roi_edge is None and cursor == Qt.SplitHCursor:
                self.unsetCursor()
                return
            elif roi_edge == 0 or roi_edge == 1:
                self.setCursor(Qt.SplitHCursor)

        if event.buttons() == Qt.LeftButton and self.selection_area.resizing:

            if self.selection_area.selected_edge == 0:
                start = self.mapToScene(event.pos())
                self.set_selection_area(x=start.x())
                self.scroll_mouse(start.x())
                return

            if self.selection_area.selected_edge == 1:
                start = QPoint(self.selection_area.rect().x(), self.selection_area.rect().y())
                end = self.mapToScene(event.pos())

                self.set_selection_area(w=end.x() - start.x())
                self.scroll_mouse(end.x())
                return

        if self.mouse_press_pos is None:
            return

        self.mouse_pos = event.pos()
        if event.buttons() == Qt.LeftButton and not self.selection_area.finished:
            start = self.mapToScene(self.mouse_press_pos)
            end = self.mapToScene(self.mouse_pos)
            self.set_selection_area(w=end.x() - start.x())
            self.scroll_mouse(end.x())

    def scroll_mouse(self, mouse_x: int):
        """
        Scrolls the mouse if ROI Selection reaches corner of view

        :param mouse_x:
        :return:
        """
        scrollbar = self.horizontalScrollBar()

        if mouse_x - self.view_rect().x() > self.view_rect().width():
            scrollbar.setValue(scrollbar.value() + 5)

        elif mouse_x < self.view_rect().x():
            scrollbar.setValue(scrollbar.value() - 5)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if self.scene() is None:
            return

        cursor = self.cursor().shape()
        if cursor == Qt.ClosedHandCursor:
            self.grab_start = None
            self.setCursor(Qt.OpenHandCursor)

        elif self.separation_area_moving:
            y_sep = self.mapToScene(event.pos()).y()
            y = self.sceneRect().y()
            h = self.sceneRect().height()
            if y_sep < y:
                y_sep = y
            elif y_sep > y + h:
                y_sep = y + h

            self.scene().draw_sep_area(y_sep)
            self.sep_area_moving.emit(y_sep)
            self.separation_area_moving = False
            self.y_sep = y_sep
            self.sep_area_changed.emit(-y_sep)
            self.unsetCursor()

        self.selection_area.finished = True
        self.selection_area.resizing = False
        self.emit_sel_area_width_changed()
        self.sel_area_start_end_changed.emit(self.selection_area.start, self.selection_area.end)

    def set_selection_area(self, x=None, w=None):
        self.selection_area.setY(self.view_rect().y())
        self.selection_area.height = self.view_rect().height()

        if x is not None:
            if x < self.sceneRect().x():
                x = self.sceneRect().x()
            elif x > self.sceneRect().x() + self.sceneRect().width():
                x = self.sceneRect().x() + self.sceneRect().width()

            self.selection_area.setX(x)

        if w is not None:
            x = self.selection_area.x
            if x + w < self.sceneRect().x():
                w = self.sceneRect().x() - x
            elif x + w > self.sceneRect().x() + self.sceneRect().width():
                w = (self.sceneRect().x() + self.sceneRect().width()) - x

            self.selection_area.width = w

        self.emit_sel_area_width_changed()

    def emit_sel_area_width_changed(self):
        self.sel_area_width_changed.emit(int(self.selection_area.width))

    def view_rect(self) -> QRectF:
        """
        Return the boundaries of the view in scene coordinates
        """
        top_left = self.mapToScene(0, 0)
        bottom_right = self.mapToScene(self.viewport().width() - 1, self.viewport().height() - 1)
        return QRectF(top_left, bottom_right)



cimport chackrf
from libc.stdlib cimport malloc
from libc.string cimport memcpy
import time

cdef object f
from cpython cimport PyBytes_GET_SIZE

cdef int _c_callback_recv(chackrf.hackrf_transfer*transfer)  with gil:
    global f
    (<object> f)(transfer.buffer[0:transfer.valid_length])
    return 0

cdef int _c_callback_send(chackrf.hackrf_transfer*transfer)  with gil:
    global f
    cdef bytes bytebuf = (<object> f)(transfer.valid_length)
    memcpy(transfer.buffer, <void*> bytebuf, PyBytes_GET_SIZE(bytebuf))
    return 0

cdef chackrf.hackrf_device*_c_device
cdef int hackrf_success = chackrf.HACKRF_SUCCESS

cpdef setup():
    chackrf.hackrf_init()
    return open()

cpdef exit():
    return chackrf.hackrf_exit()

cpdef reopen():
    close()
    return open()

cpdef open():
    return chackrf.hackrf_open(&_c_device)

cpdef close():
    time.sleep(0.1)
    return chackrf.hackrf_close(_c_device)

cpdef start_rx_mode(callback):
    global f
    f = callback
    return chackrf.hackrf_start_rx(_c_device, _c_callback_recv, <void*> _c_callback_recv)

cpdef stop_rx_mode():
    return chackrf.hackrf_stop_rx(_c_device)

cpdef start_tx_mode(callback):
    global f
    f = callback
    return chackrf.hackrf_start_tx(_c_device, _c_callback_send, <void *> _c_callback_send)

cpdef stop_tx_mode():
    return chackrf.hackrf_stop_tx(_c_device)

cpdef board_id_read():
    cdef unsigned char value
    ret = chackrf.hackrf_board_id_read(_c_device, &value)
    if ret == hackrf_success:
        return value
    else:
        return ""

cpdef version_string_read():
    cdef char*version = <char *> malloc(20 * sizeof(char))
    cdef unsigned char length = 20
    ret = chackrf.hackrf_version_string_read(_c_device, version, length)
    if ret == hackrf_success:
        return version.decode('UTF-8')
    else:
        return ""

cpdef set_freq(freq_hz):
    return chackrf.hackrf_set_freq(_c_device, freq_hz)

cpdef is_streaming():
    ret = chackrf.hackrf_is_streaming(_c_device)
    if ret == 1:
        return True
    else:
        return False

cpdef set_rf_gain(value):
    """ Enable or disable RF amplifier """
    return set_amp_enable(value)

cpdef set_if_rx_gain(value):
    """ Sets the LNA gain, in 8Db steps, maximum value of 40 """
    return chackrf.hackrf_set_lna_gain(_c_device, value)

cpdef set_if_tx_gain(value):
    """ Sets the txvga gain, in 1db steps, maximum value of 47 """
    return chackrf.hackrf_set_txvga_gain(_c_device, value)

cpdef set_baseband_gain(value):
    """ Sets the vga gain, in 2db steps, maximum value of 62 """
    return chackrf.hackrf_set_vga_gain(_c_device, value)

cpdef set_antenna_enable(value):
    cdef bint val = 1 if value else 0
    return chackrf.hackrf_set_antenna_enable(_c_device, val)

cpdef set_sample_rate(sample_rate):
    return chackrf.hackrf_set_sample_rate(_c_device, sample_rate)

cpdef set_amp_enable(value):
    cdef bint val = 1 if value else 0
    return chackrf.hackrf_set_amp_enable(_c_device, val)

cpdef set_baseband_filter_bandwidth(bandwidth_hz):
    return chackrf.hackrf_set_baseband_filter_bandwidth(_c_device, bandwidth_hz)




import numpy as np
import psutil
from urh import constants
import zmq
from PyQt5.QtCore import pyqtSignal

from urh.dev.gr.AbstractBaseThread import AbstractBaseThread
from urh.util.Logger import logger


class ReceiverThread(AbstractBaseThread):
    index_changed = pyqtSignal(int, int)

    def __init__(self, freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, ip='127.0.0.1',
                 parent=None,  is_ringbuffer=False):
        super().__init__(freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, True, ip, parent)

        self.is_ringbuffer = is_ringbuffer  # Ringbuffer for Live Sniffing
        self.data = None

    def init_recv_buffer(self):
        # Take 60% of free memory
        nsamples = int(constants.SETTINGS.value('ram_threshold', 0.6, float) * (psutil.virtual_memory().available / 8))
        self.data = np.zeros(nsamples, dtype=np.complex64)

    def run(self):
        if self.data is None:
            self.init_recv_buffer()

        self.initialize_process()
        logger.info("Initialize receive socket")
        self.init_recv_socket()

        recv = self.socket.recv
        rcvd = b""

        try:
            while not self.isInterruptionRequested():

                try:
                    rcvd += recv(32768)  # Receive Buffer = 32768 Byte
                except (zmq.error.ContextTerminated, ConnectionResetError):
                    self.stop("Stopped receiving, because connection was reset.")
                    return
                except OSError as e:   # https://github.com/jopohl/urh/issues/131
                    logger.warning("Error occurred", str(e))

                if len(rcvd) < 8:
                    self.stop("Stopped receiving: No data received anymore")
                    return

                if len(rcvd) % 8 != 0:
                    continue

                try:
                    tmp = np.fromstring(rcvd, dtype=np.complex64)

                    len_tmp = len(tmp)
                    if self.data is None:
                        # seems to be sometimes None in rare cases
                        self.init_recv_buffer()

                    if self.current_index + len_tmp >= len(self.data):
                        if self.is_ringbuffer:
                            self.current_index = 0
                            if len_tmp >= len(self.data):
                                self.stop("Receiving buffer too small.")
                        else:
                            self.stop("Receiving Buffer is full.")
                            return
                    self.data[self.current_index:self.current_index + len_tmp] = tmp
                    self.current_index += len_tmp
                    self.index_changed.emit(self.current_index - len_tmp,
                                            self.current_index)

                    rcvd = b""
                except ValueError:
                    self.stop("Could not receive data. Is your Hardware ok?")
        except RuntimeError:
            logger.error("Receiver Thread crashed.")



import numpy as np
from urh.dev.native.Device import Device
from urh.dev.native.lib import hackrf
from urh.util.Logger import logger


class HackRF(Device):
    BYTES_PER_SAMPLE = 2  # HackRF device produces 8 bit unsigned IQ data

    @staticmethod
    def initialize_hackrf(freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, ctrl_conn, is_tx):
        ret = hackrf.setup()
        ctrl_conn.send("setup:" + str(ret))
        if ret != 0:
            return False

        ret = hackrf.set_freq(freq)
        ctrl_conn.send("set_center_freq to {0}:{1}".format(freq, ret))

        ret = hackrf.set_sample_rate(sample_rate)
        ctrl_conn.send("set_sample_rate to {0}:{1}".format(sample_rate, ret))

        ret = hackrf.set_rf_gain(gain)
        ctrl_conn.send("set_rf_gain to {0}:{1}".format(gain, ret))

        if is_tx:
            ret = hackrf.set_if_tx_gain(if_gain)
            ctrl_conn.send("set_if_gain to {0}:{1}".format(if_gain, ret))
        else:
            ret = hackrf.set_if_rx_gain(if_gain)
            ctrl_conn.send("set_if_gain to {0}:{1}".format(if_gain, ret))

            ret = hackrf.set_baseband_gain(baseband_gain)
            ctrl_conn.send("set_baseband_gain to {0}:{1}".format(baseband_gain, ret))

        ret = hackrf.set_baseband_filter_bandwidth(bandwidth)
        ctrl_conn.send("set_bandwidth to {0}:{1}".format(bandwidth, ret))

        return True

    @staticmethod
    def shutdown_hackrf(ctrl_conn):
        logger.debug("HackRF: closing device")
        ret = hackrf.close()
        ctrl_conn.send("close:" + str(ret))

        ret = hackrf.exit()
        ctrl_conn.send("exit:" + str(ret))

        return True

    @staticmethod
    def hackrf_receive(data_connection, ctrl_connection, freq, sample_rate, bandwidth, gain, if_gain, baseband_gain):
        def callback_recv(buffer):
            try:
                data_connection.send_bytes(buffer)
            except (BrokenPipeError, EOFError):
                pass
            return 0

        if not HackRF.initialize_hackrf(freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, ctrl_connection, is_tx=False):
            return False

        hackrf.start_rx_mode(callback_recv)

        exit_requested = False

        while not exit_requested:
            while ctrl_connection.poll():
                result = HackRF.process_command(ctrl_connection.recv(), is_tx=False)
                if result == "stop":
                    exit_requested = True
                    break

        HackRF.shutdown_hackrf(ctrl_connection)
        data_connection.close()
        ctrl_connection.close()

    @staticmethod
    def hackrf_send(ctrl_connection, freq, sample_rate, bandwidth, gain, if_gain, baseband_gain,
                    send_buffer, current_sent_index, current_sending_repeat, sending_repeats):
        def sending_is_finished():
            if sending_repeats == 0:  # 0 = infinity
                return False

            return current_sending_repeat.value >= sending_repeats and current_sent_index.value >= len(send_buffer)

        def callback_send(buffer_length):
            try:
                if sending_is_finished():
                    return b""

                result = send_buffer[current_sent_index.value:current_sent_index.value + buffer_length]
                current_sent_index.value += buffer_length
                if current_sent_index.value >= len(send_buffer) - 1:
                    current_sending_repeat.value += 1
                    if current_sending_repeat.value < sending_repeats or sending_repeats == 0:  # 0 = infinity
                        current_sent_index.value = 0
                    else:
                        current_sent_index.value = len(send_buffer)

                return result
            except (BrokenPipeError, EOFError):
                return b""

        if not HackRF.initialize_hackrf(freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, ctrl_connection, is_tx=True):
            return False

        hackrf.start_tx_mode(callback_send)

        exit_requested = False

        while not exit_requested and not sending_is_finished():
            while ctrl_connection.poll():
                result = HackRF.process_command(ctrl_connection.recv(), is_tx=True)
                if result == "stop":
                    exit_requested = True
                    break

        if exit_requested:
            logger.debug("HackRF: exit requested. Stopping sending")
        if sending_is_finished():
            logger.debug("HackRF: sending is finished.")

        HackRF.shutdown_hackrf(ctrl_connection)
        ctrl_connection.close()

    @staticmethod
    def process_command(command, is_tx: bool):
        is_rx = not is_tx
        if command == "stop":
            return "stop"

        tag, value = command.split(":")
        if tag == "center_freq":
            logger.info("HackRF: Set center freq to {0}".format(int(value)))
            return hackrf.set_freq(int(value))

        elif tag == "rf_gain":
            logger.info("HackRF: Set gain to {0}".format(int(value)))
            hackrf.set_rf_gain(int(value))

        elif tag == "if_gain" and is_rx:
            logger.info("HackRF: Set if gain to {0}".format(int(value)))
            hackrf.set_if_rx_gain(int(value))

        elif tag == "if_gain" and is_tx:
            logger.info("HackRF: Set if gain to {0}".format(int(value)))
            hackrf.set_if_tx_gain(int(value))

        elif tag == "baseband_gain" and is_rx:
            logger.info("HackRF: Set baseband gain to {0}".format(int(value)))
            hackrf.set_baseband_gain(int(value))

        elif tag == "sample_rate":
            logger.info("HackRF: Set sample_rate to {0}".format(int(value)))
            return hackrf.set_sample_rate(int(value))

        elif tag == "bandwidth":
            logger.info("HackRF: Set bandwidth to {0}".format(int(value)))
            return hackrf.set_baseband_filter_bandwidth(int(value))

    def __init__(self, center_freq, sample_rate, bandwidth, gain, if_gain=1, baseband_gain=1, is_ringbuffer=False):
        super().__init__(center_freq=center_freq, sample_rate=sample_rate, bandwidth=bandwidth,
                         gain=gain, if_gain=if_gain, baseband_gain=baseband_gain, is_ringbuffer=is_ringbuffer)
        self.success = 0

        self.receive_process_function = HackRF.hackrf_receive
        self.send_process_function = HackRF.hackrf_send

        self.error_codes = {
            0: "HACKRF_SUCCESS",
            1: "HACKRF_TRUE",
            1337: "TIMEOUT ERROR",
            -2: "HACKRF_ERROR_INVALID_PARAM",
            -5: "HACKRF_ERROR_NOT_FOUND",
            -6: "HACKRF_ERROR_BUSY",
            -11: "HACKRF_ERROR_NO_MEM",
            -1000: "HACKRF_ERROR_LIBUSB",
            -1001: "HACKRF_ERROR_THREAD",
            -1002: "HACKRF_ERROR_STREAMING_THREAD_ERR",
            -1003: "HACKRF_ERROR_STREAMING_STOPPED",
            -1004: "HACKRF_ERROR_STREAMING_EXIT_CALLED",
            -4242: "HACKRF NOT OPEN",
            -9999: "HACKRF_ERROR_OTHER"
        }

    @staticmethod
    def unpack_complex(buffer, nvalues: int):
        result = np.empty(nvalues, dtype=np.complex64)
        unpacked = np.frombuffer(buffer, dtype=[('r', np.int8), ('i', np.int8)])
        result.real = (unpacked['r'] + 0.5) / 127.5
        result.imag = (unpacked['i'] + 0.5) / 127.5
        return result

    @staticmethod
    def pack_complex(complex_samples: np.ndarray):
        assert complex_samples.dtype == np.complex64
        # tostring() is a compatibility (numpy<1.9) alias for tobytes(). Despite its name it returns bytes not strings.
        return (127.5 * ((complex_samples.view(np.float32)) - 0.5 / 127.5)).astype(np.int8).tostring()



import math
from PyQt5.QtCore import QAbstractTableModel, pyqtSignal, Qt, QModelIndex
from PyQt5.QtGui import QFont

from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Message import Message
from urh.signalprocessing.ProtocolGroup import ProtocolGroup


class PLabelTableModel(QAbstractTableModel):
    header_labels = ["Name", "Start", "End", 'Color', 'Apply decoding']

    label_removed = pyqtSignal(ProtocolLabel)
    apply_decoding_changed = pyqtSignal(ProtocolLabel)

    def __init__(self, message: Message, field_types, parent=None):
        """

        :param message:
        :type field_types: list of FieldType
        :param parent:
        """
        super().__init__(parent)
        self.row_count = len(message.message_type)
        self.proto_view = 0
        self.message = message
        self.message_type = message.message_type
        self.field_types_by_caption = {ft.caption: ft for ft in field_types}
        self.update()

    def update(self):
        self.beginResetModel()
        self.endResetModel()

    def columnCount(self, parent: QModelIndex=None, *args, **kwargs):
        return len(self.header_labels)

    def rowCount(self, parent: QModelIndex=None, *args, **kwargs):
        return len(self.message_type)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        return super().headerData(section, orientation, role)

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        i, j = index.row(), index.column()
        if role == Qt.DisplayRole:
            lbl = self.message_type[i]
            if j == 0:
                return lbl.name
            elif j == 1:
                return self.message.get_label_range(lbl, view=self.proto_view, decode=True)[0] + 1
            elif j == 2:
                return self.message.get_label_range(lbl, view=self.proto_view, decode=True)[1]
            elif j == 3:
                return lbl.color_index
            elif j == 4:
                return lbl.apply_decoding
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignCenter
        elif role == Qt.FontRole and j == 0:
            font = QFont()
            font.setItalic(self.message_type[i].type is None)
            return font
        else:
            return None

    def setData(self, index: QModelIndex, value, role=Qt.EditRole):
        if value == "":
            return True

        i = index.row()
        j = index.column()
        if i >= len(self.message_type):
            return False

        lbl = self.message_type[i]

        if j == 0:
            lbl.name = value
            if value in self.field_types_by_caption:
                lbl.type = self.field_types_by_caption[value]
            else:
                lbl.type = None
        elif j == 1:
            lbl.start = self.message.convert_index(int(value-1), from_view=self.proto_view, to_view=0, decoded=True)[0]
        elif j == 2:
            lbl.end = self.message.convert_index(int(value), from_view=self.proto_view, to_view=0, decoded=True)[0]
        elif j == 3:
            lbl.color_index = value
        elif j == 4:
            if bool(value) != lbl.apply_decoding:
                lbl.apply_decoding = bool(value)
                self.apply_decoding_changed.emit(lbl)

        return True

    def flags(self, index):
        if not index.isValid():
            return Qt.NoItemFlags

        try:
            _ = self.message_type[index.row()]
        except IndexError:
            return Qt.NoItemFlags

        return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def remove_label(self, label):
        self.message_type.remove(label)
        self.update()
        self.label_removed.emit(label)

    def remove_label_at(self, index: int):
        try:
            label = self.message_type[index]
            self.remove_label(label)
        except IndexError:
            pass



import select
import socket

import numpy
import numpy as np
import time
import zmq

from urh.dev.gr.AbstractBaseThread import AbstractBaseThread
from urh.util.Logger import logger


class SenderThread(AbstractBaseThread):
    MAX_SAMPLES_PER_TRANSMISSION = 65536

    def __init__(self, freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, ip='127.0.0.1', parent=None):
        super().__init__(freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, False, ip, parent)

        self.data = numpy.empty(1, dtype=numpy.complex64)
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUSH)
        self.gr_port = self.socket.bind_to_random_port("tcp://{0}".format(self.ip))
        self.max_repeats = 1  # How often shall we send the data?

        self.__samples_per_transmission = self.MAX_SAMPLES_PER_TRANSMISSION

    @property
    def repeat_endless(self):
        return self.max_repeats == 0 or self.max_repeats == -1

    @property
    def samples_per_transmission(self):
        return self.__samples_per_transmission

    @samples_per_transmission.setter
    def samples_per_transmission(self, val: int):
        if val >= self.MAX_SAMPLES_PER_TRANSMISSION:
            self.__samples_per_transmission = self.MAX_SAMPLES_PER_TRANSMISSION
        elif val <= 1:
            self.__samples_per_transmission = 1
        else:
            self.__samples_per_transmission = 2 ** (int(np.log2(val)) - 1)

    def run(self):
        self.initialize_process()
        len_data = len(self.data)
        self.current_iteration = self.current_iteration if self.current_iteration is not None else 0
        time.sleep(1)

        try:
            while self.current_index < len_data and not self.isInterruptionRequested():
                time.sleep(0.1 * (self.samples_per_transmission / self.MAX_SAMPLES_PER_TRANSMISSION))
                self.socket.send(self.data[self.current_index:self.current_index + self.samples_per_transmission].tostring())
                self.current_index += self.samples_per_transmission

                if self.current_index >= len_data:
                    self.current_iteration += 1
                else:
                    continue

                if self.repeat_endless or self.current_iteration < self.max_repeats:
                    self.current_index = 0

            self.current_index = len_data - 1
            self.current_iteration = None
            self.stop("FIN - All data was sent successfully")
        except RuntimeError:
            logger.error("Sender thread crashed.")



# -*- coding: utf-8 -*-

#
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1017, 884)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/data/icons/appicon.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        MainWindow.setTabShape(QtWidgets.QTabWidget.Rounded)
        MainWindow.setDockNestingEnabled(False)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.splitter = QtWidgets.QSplitter(self.centralwidget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setChildrenCollapsible(True)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.lnEdtTreeFilter = QtWidgets.QLineEdit(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lnEdtTreeFilter.sizePolicy().hasHeightForWidth())
        self.lnEdtTreeFilter.setSizePolicy(sizePolicy)
        self.lnEdtTreeFilter.setAcceptDrops(False)
        self.lnEdtTreeFilter.setInputMethodHints(QtCore.Qt.ImhDialableCharactersOnly)
        self.lnEdtTreeFilter.setClearButtonEnabled(True)
        self.lnEdtTreeFilter.setObjectName("lnEdtTreeFilter")
        self.verticalLayout_3.addWidget(self.lnEdtTreeFilter)
        self.fileTree = DirectoryTreeView(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(10)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fileTree.sizePolicy().hasHeightForWidth())
        self.fileTree.setSizePolicy(sizePolicy)
        self.fileTree.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.fileTree.setAutoScroll(True)
        self.fileTree.setDragEnabled(True)
        self.fileTree.setDragDropMode(QtWidgets.QAbstractItemView.DragOnly)
        self.fileTree.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.fileTree.setSortingEnabled(False)
        self.fileTree.setObjectName("fileTree")
        self.fileTree.header().setCascadingSectionResizes(True)
        self.fileTree.header().setStretchLastSection(False)
        self.verticalLayout_3.addWidget(self.fileTree)
        self.tabWidget_Project = QtWidgets.QTabWidget(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget_Project.sizePolicy().hasHeightForWidth())
        self.tabWidget_Project.setSizePolicy(sizePolicy)
        self.tabWidget_Project.setObjectName("tabWidget_Project")
        self.tabParticipants = QtWidgets.QWidget()
        self.tabParticipants.setObjectName("tabParticipants")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.tabParticipants)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.listViewParticipants = QtWidgets.QListView(self.tabParticipants)
        self.listViewParticipants.setObjectName("listViewParticipants")
        self.horizontalLayout.addWidget(self.listViewParticipants)
        self.tabWidget_Project.addTab(self.tabParticipants, "")
        self.tabDescription = QtWidgets.QWidget()
        self.tabDescription.setObjectName("tabDescription")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.tabDescription)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.textEditProjectDescription = QtWidgets.QTextEdit(self.tabDescription)
        self.textEditProjectDescription.setObjectName("textEditProjectDescription")
        self.horizontalLayout_2.addWidget(self.textEditProjectDescription)
        self.tabWidget_Project.addTab(self.tabDescription, "")
        self.verticalLayout_3.addWidget(self.tabWidget_Project)
        self.verticalLayout_3.setStretch(1, 3)
        self.tabWidget = QtWidgets.QTabWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_interpretation = QtWidgets.QWidget()
        self.tab_interpretation.setObjectName("tab_interpretation")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab_interpretation)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tabWidget.addTab(self.tab_interpretation, "")
        self.tab_protocol = QtWidgets.QWidget()
        self.tab_protocol.setObjectName("tab_protocol")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.tab_protocol)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget.addTab(self.tab_protocol, "")
        self.tab_generator = QtWidgets.QWidget()
        self.tab_generator.setObjectName("tab_generator")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.tab_generator)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.tabWidget.addTab(self.tab_generator, "")
        self.verticalLayout_4.addWidget(self.splitter)
        self.progressBar = QtWidgets.QProgressBar(self.centralwidget)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_4.addWidget(self.progressBar)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1017, 21))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuEdit = QtWidgets.QMenu(self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionFSK = QtWidgets.QAction(MainWindow)
        self.actionFSK.setCheckable(True)
        self.actionFSK.setObjectName("actionFSK")
        self.actionOOK = QtWidgets.QAction(MainWindow)
        self.actionOOK.setCheckable(True)
        self.actionOOK.setChecked(True)
        self.actionOOK.setObjectName("actionOOK")
        self.actionPSK = QtWidgets.QAction(MainWindow)
        self.actionPSK.setCheckable(True)
        self.actionPSK.setObjectName("actionPSK")
        self.actionNone = QtWidgets.QAction(MainWindow)
        self.actionNone.setCheckable(True)
        self.actionNone.setObjectName("actionNone")
        self.actionAuto_Fit_Y = QtWidgets.QAction(MainWindow)
        self.actionAuto_Fit_Y.setCheckable(True)
        self.actionAuto_Fit_Y.setChecked(True)
        self.actionAuto_Fit_Y.setObjectName("actionAuto_Fit_Y")
        self.actionUndo = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("edit-undo")
        self.actionUndo.setIcon(icon)
        self.actionUndo.setObjectName("actionUndo")
        self.actionRedo = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("edit-redo")
        self.actionRedo.setIcon(icon)
        self.actionRedo.setObjectName("actionRedo")
        self.actionShow_Confirm_Close_Dialog = QtWidgets.QAction(MainWindow)
        self.actionShow_Confirm_Close_Dialog.setCheckable(True)
        self.actionShow_Confirm_Close_Dialog.setChecked(False)
        self.actionShow_Confirm_Close_Dialog.setObjectName("actionShow_Confirm_Close_Dialog")
        self.actionTest = QtWidgets.QAction(MainWindow)
        self.actionTest.setObjectName("actionTest")
        self.actionHold_Shift_to_Drag = QtWidgets.QAction(MainWindow)
        self.actionHold_Shift_to_Drag.setCheckable(True)
        self.actionHold_Shift_to_Drag.setChecked(False)
        self.actionHold_Shift_to_Drag.setObjectName("actionHold_Shift_to_Drag")
        self.actionDocumentation = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("help-contents")
        self.actionDocumentation.setIcon(icon)
        self.actionDocumentation.setIconVisibleInMenu(True)
        self.actionDocumentation.setObjectName("actionDocumentation")
        self.actionAbout_AutomaticHacker = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("help-about")
        self.actionAbout_AutomaticHacker.setIcon(icon)
        self.actionAbout_AutomaticHacker.setIconVisibleInMenu(True)
        self.actionAbout_AutomaticHacker.setObjectName("actionAbout_AutomaticHacker")
        self.actionOpenSignal = QtWidgets.QAction(MainWindow)
        self.actionOpenSignal.setObjectName("actionOpenSignal")
        self.actionOpenProtocol = QtWidgets.QAction(MainWindow)
        self.actionOpenProtocol.setObjectName("actionOpenProtocol")
        self.actionShow_Compare_Frame = QtWidgets.QAction(MainWindow)
        self.actionShow_Compare_Frame.setCheckable(True)
        self.actionShow_Compare_Frame.setChecked(True)
        self.actionShow_Compare_Frame.setObjectName("actionShow_Compare_Frame")
        self.actionClose_all = QtWidgets.QAction(MainWindow)
        self.actionClose_all.setIconVisibleInMenu(True)
        self.actionClose_all.setObjectName("actionClose_all")
        self.actionMinimize_all = QtWidgets.QAction(MainWindow)
        self.actionMinimize_all.setIconVisibleInMenu(True)
        self.actionMinimize_all.setObjectName("actionMinimize_all")
        self.actionMaximize_all = QtWidgets.QAction(MainWindow)
        self.actionMaximize_all.setIconVisibleInMenu(True)
        self.actionMaximize_all.setObjectName("actionMaximize_all")
        self.actionSaveAllSignals = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("document-save")
        self.actionSaveAllSignals.setIcon(icon)
        self.actionSaveAllSignals.setIconVisibleInMenu(True)
        self.actionSaveAllSignals.setObjectName("actionSaveAllSignals")
        self.actionSeperate_Protocols_in_Compare_Frame = QtWidgets.QAction(MainWindow)
        self.actionSeperate_Protocols_in_Compare_Frame.setCheckable(True)
        self.actionSeperate_Protocols_in_Compare_Frame.setChecked(True)
        self.actionSeperate_Protocols_in_Compare_Frame.setObjectName("actionSeperate_Protocols_in_Compare_Frame")
        self.actionOpenArchive = QtWidgets.QAction(MainWindow)
        self.actionOpenArchive.setObjectName("actionOpenArchive")
        self.actionOpen = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("document-open")
        self.actionOpen.setIcon(icon)
        self.actionOpen.setIconVisibleInMenu(True)
        self.actionOpen.setObjectName("actionOpen")
        self.actionOpen_Folder = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("folder-open")
        self.actionOpen_Folder.setIcon(icon)
        self.actionOpen_Folder.setObjectName("actionOpen_Folder")
        self.actionShow_only_Compare_Frame = QtWidgets.QAction(MainWindow)
        self.actionShow_only_Compare_Frame.setCheckable(True)
        self.actionShow_only_Compare_Frame.setChecked(True)
        self.actionShow_only_Compare_Frame.setObjectName("actionShow_only_Compare_Frame")
        self.actionConfigurePlugins = QtWidgets.QAction(MainWindow)
        self.actionConfigurePlugins.setIconVisibleInMenu(True)
        self.actionConfigurePlugins.setObjectName("actionConfigurePlugins")
        self.actionSort_Frames_by_Name = QtWidgets.QAction(MainWindow)
        self.actionSort_Frames_by_Name.setObjectName("actionSort_Frames_by_Name")
        self.actionConvert_Folder_to_Project = QtWidgets.QAction(MainWindow)
        self.actionConvert_Folder_to_Project.setIconVisibleInMenu(True)
        self.actionConvert_Folder_to_Project.setObjectName("actionConvert_Folder_to_Project")
        self.actionDecoding = QtWidgets.QAction(MainWindow)
        self.actionDecoding.setObjectName("actionDecoding")
        self.actionRecord = QtWidgets.QAction(MainWindow)
        self.actionRecord.setIconVisibleInMenu(True)
        self.actionRecord.setObjectName("actionRecord")
        self.actionSpectrum_Analyzer = QtWidgets.QAction(MainWindow)
        self.actionSpectrum_Analyzer.setIconVisibleInMenu(True)
        self.actionSpectrum_Analyzer.setObjectName("actionSpectrum_Analyzer")
        self.actionOptions = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("configure")
        self.actionOptions.setIcon(icon)
        self.actionOptions.setIconVisibleInMenu(True)
        self.actionOptions.setObjectName("actionOptions")
        self.actionShow_file_tree = QtWidgets.QAction(MainWindow)
        self.actionShow_file_tree.setCheckable(True)
        self.actionShow_file_tree.setObjectName("actionShow_file_tree")
        self.actionNew_Project = QtWidgets.QAction(MainWindow)
        self.actionNew_Project.setObjectName("actionNew_Project")
        self.actionSniff_protocol = QtWidgets.QAction(MainWindow)
        self.actionSniff_protocol.setObjectName("actionSniff_protocol")
        self.actionProject_settings = QtWidgets.QAction(MainWindow)
        self.actionProject_settings.setObjectName("actionProject_settings")
        self.actionSave_project = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("document-save")
        self.actionSave_project.setIcon(icon)
        self.actionSave_project.setObjectName("actionSave_project")
        self.actionFullscreen_mode = QtWidgets.QAction(MainWindow)
        self.actionFullscreen_mode.setCheckable(True)
        self.actionFullscreen_mode.setObjectName("actionFullscreen_mode")
        self.actionOpen_directory = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon.fromTheme("folder-open")
        self.actionOpen_directory.setIcon(icon)
        self.actionOpen_directory.setObjectName("actionOpen_directory")
        self.actionAbout_Qt = QtWidgets.QAction(MainWindow)
        self.actionAbout_Qt.setObjectName("actionAbout_Qt")
        self.menuFile.addAction(self.actionNew_Project)
        self.menuFile.addAction(self.actionProject_settings)
        self.menuFile.addAction(self.actionSave_project)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addAction(self.actionOpen_directory)
        self.menuFile.addAction(self.actionRecord)
        self.menuFile.addAction(self.actionSniff_protocol)
        self.menuFile.addAction(self.actionSpectrum_Analyzer)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSaveAllSignals)
        self.menuFile.addAction(self.actionClose_all)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionConvert_Folder_to_Project)
        self.menuEdit.addAction(self.actionMinimize_all)
        self.menuEdit.addAction(self.actionMaximize_all)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionDecoding)
        self.menuEdit.addAction(self.actionOptions)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionFullscreen_mode)
        self.menuHelp.addAction(self.actionAbout_AutomaticHacker)
        self.menuHelp.addAction(self.actionAbout_Qt)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget_Project.setCurrentIndex(0)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Universal Radio Hacker"))
        self.lnEdtTreeFilter.setPlaceholderText(_translate("MainWindow", "Filter"))
        self.tabWidget_Project.setTabText(self.tabWidget_Project.indexOf(self.tabParticipants), _translate("MainWindow", "Participants"))
        self.tabWidget_Project.setTabText(self.tabWidget_Project.indexOf(self.tabDescription), _translate("MainWindow", "Description"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_interpretation), _translate("MainWindow", "Interpretation"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_protocol), _translate("MainWindow", "Analysis"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_generator), _translate("MainWindow", "Generator"))
        self.menuFile.setTitle(_translate("MainWindow", "Fi&le"))
        self.menuEdit.setTitle(_translate("MainWindow", "Edi&t"))
        self.menuHelp.setTitle(_translate("MainWindow", "Hel&p"))
        self.actionFSK.setText(_translate("MainWindow", "Undo"))
        self.actionOOK.setText(_translate("MainWindow", "Redo"))
        self.actionPSK.setText(_translate("MainWindow", "PSK"))
        self.actionNone.setText(_translate("MainWindow", "None (bei .bin)"))
        self.actionAuto_Fit_Y.setText(_translate("MainWindow", "&Auto Fit Y"))
        self.actionUndo.setText(_translate("MainWindow", "&Undo"))
        self.actionRedo.setText(_translate("MainWindow", "&Redo"))
        self.actionShow_Confirm_Close_Dialog.setText(_translate("MainWindow", "&Show Confirm Close Dialog"))
        self.actionTest.setText(_translate("MainWindow", "test"))
        self.actionHold_Shift_to_Drag.setText(_translate("MainWindow", "&Hold Shift to Drag"))
        self.actionDocumentation.setText(_translate("MainWindow", "&Documentation"))
        self.actionAbout_AutomaticHacker.setText(_translate("MainWindow", "&About Universal Radio Hacker..."))
        self.actionOpenSignal.setText(_translate("MainWindow", "&Signal"))
        self.actionOpenProtocol.setText(_translate("MainWindow", "&Protocol"))
        self.actionShow_Compare_Frame.setText(_translate("MainWindow", "Show &Compare Frame"))
        self.actionClose_all.setText(_translate("MainWindow", "&Close all"))
        self.actionMinimize_all.setText(_translate("MainWindow", "&Minimize all"))
        self.actionMaximize_all.setText(_translate("MainWindow", "Maximize &all"))
        self.actionSaveAllSignals.setText(_translate("MainWindow", "&Save all signals"))
        self.actionSeperate_Protocols_in_Compare_Frame.setText(_translate("MainWindow", "Seperate &Protocols in Compare Frame"))
        self.actionOpenArchive.setText(_translate("MainWindow", "&Archive"))
        self.actionOpen.setText(_translate("MainWindow", "&Open..."))
        self.actionOpen_Folder.setText(_translate("MainWindow", "Open &Folder.."))
        self.actionShow_only_Compare_Frame.setText(_translate("MainWindow", "Show Compare Frame only"))
        self.actionConfigurePlugins.setText(_translate("MainWindow", "Configure..."))
        self.actionSort_Frames_by_Name.setText(_translate("MainWindow", "Sort &Frames by Name"))
        self.actionConvert_Folder_to_Project.setText(_translate("MainWindow", "Convert &Folder to Project"))
        self.actionDecoding.setText(_translate("MainWindow", "&Decoding..."))
        self.actionRecord.setText(_translate("MainWindow", "&Record signal..."))
        self.actionSpectrum_Analyzer.setText(_translate("MainWindow", "Spectrum &Analyzer..."))
        self.actionOptions.setText(_translate("MainWindow", "&Options..."))
        self.actionShow_file_tree.setText(_translate("MainWindow", "Sh&ow file tree"))
        self.actionNew_Project.setText(_translate("MainWindow", "&New Project.."))
        self.actionSniff_protocol.setText(_translate("MainWindow", "Sn&iff protocol..."))
        self.actionProject_settings.setText(_translate("MainWindow", "&Project settings..."))
        self.actionSave_project.setText(_translate("MainWindow", "Sa&ve project"))
        self.actionFullscreen_mode.setText(_translate("MainWindow", "&Fullscreen mode"))
        self.actionOpen_directory.setText(_translate("MainWindow", "Open &directory..."))
        self.actionAbout_Qt.setText(_translate("MainWindow", "About Qt"))

from urh.ui.views.DirectoryTreeView import DirectoryTreeView
from . import urh_rc



# noinspection PyUnresolvedReferences
from libcpp cimport bool

# noinspection PyUnresolvedReferences
import numpy as np
cimport numpy as np
from urh.cythonext import util

from cython.parallel import prange
from libc.math cimport atan2, sqrt, M_PI, sin, cos

cdef:
    float complex imag_unit = 1j

cdef float NOISE_FSK_PSK = -4.0
cdef float NOISE_ASK = 0.0


cdef float calc_costa_alpha(float bw, float damp=1 / sqrt(2)) nogil:
    # BW in range((2pi/200), (2pi/100))
    cdef float alpha = (4 * damp * bw) / (1 + 2 * damp * bw + bw * bw)

    return alpha

cdef float calc_costa_beta(float bw, float damp=1 / sqrt(2)) nogil:
    # BW in range((2pi/200), (2pi/100))
    cdef float beta = (4 * bw * bw) / (1 + 2 * damp * bw + bw * bw)
    return beta


cpdef float get_noise_for_mod_type(int mod_type):
    if mod_type == 0:
        return NOISE_ASK
    elif mod_type == 1:
        return NOISE_FSK_PSK
    elif mod_type == 2:
        return NOISE_FSK_PSK
    elif mod_type == 3:  # ASK + PSK (QAM)
        return NOISE_ASK * NOISE_FSK_PSK
    else:
        return 0


cdef void costa_demod(float complex[::1] samples, float[::1] result, float noise_sqrd,
                          float costa_alpha, float costa_beta, bool qam, long long num_samples):
    cdef float phase_error
    cdef long long i
    cdef float costa_freq = 0
    cdef float costa_phase = 0
    cdef float complex nco_out
    cdef float complex nco_times_sample, c
    cdef float real, imag
    cdef float magnitude

    for i in range(0, num_samples):
        c = samples[i]
        real, imag = c.real, c.imag
        magnitude = real * real + imag * imag
        if magnitude <= noise_sqrd:  # |c| <= mag_treshold
            result[i] = NOISE_FSK_PSK
            continue

        # # NCO Output
        #nco_out = np.exp(-costa_phase * 1j)
        nco_out = cos(-costa_phase) + imag_unit * sin(-costa_phase)

        nco_times_sample = nco_out * c
        phase_error = nco_times_sample.imag * nco_times_sample.real
        costa_freq += costa_beta * phase_error
        costa_phase += costa_freq + costa_alpha * phase_error
        if qam:
            result[i] = magnitude * nco_times_sample.real
        else:
            result[i] = nco_times_sample.real

cpdef np.ndarray[np.float32_t, ndim=1] afp_demod(float complex[::1] samples, float noise_mag, int mod_type):
    if len(samples) <= 2:
        return np.empty(len(samples), dtype=np.float32)

    cdef long long i, ns
    cdef float complex tmp = 0
    cdef float complex c = 0
    cdef float arg, noise_sqrd, complex_phase, prev_phase, NOISE
    cdef float real = 0
    cdef float imag = 0
    ns = len(samples)

    cdef float[::1] result = np.empty(ns, dtype=np.float32, order="C")
    cdef float costa_freq = 0
    cdef float costa_phase = 0
    cdef complex nco_out = 0
    cdef float phase_error
    cdef float costa_alpha, costa_beta
    cdef complex nco_times_sample
    cdef float magnitude = 0

    # Atan2 liefert Werte im Bereich von -Pi bis Pi
    # Wir nutzen die Magic Constant NOISE_FSK_PSK um Rauschen abzuschneiden
    noise_sqrd = noise_mag * noise_mag
    NOISE = get_noise_for_mod_type(mod_type)
    result[0] = NOISE

    cdef bool qam = False

    if mod_type == 2 or mod_type == 3: # PSK or QAM
        if mod_type == 3:
            qam = True

        costa_alpha = calc_costa_alpha(<float>(2 * M_PI / 100))
        costa_beta = calc_costa_beta(<float>(2 * M_PI / 100))
        costa_demod(samples, result, noise_sqrd, costa_alpha, costa_beta, qam, ns)

    else:
        for i in prange(1, ns, nogil=True, schedule='static'):
            c = samples[i]
            real, imag = c.real, c.imag
            magnitude = real * real + imag * imag
            if magnitude <= noise_sqrd:  # |c| <= mag_treshold
                result[i] = NOISE
                continue

            if mod_type == 0:  # ASK
                result[i] = magnitude
            elif mod_type == 1:  # FSK
                tmp = samples[i - 1].conjugate() * c
                result[i] = atan2(tmp.imag, tmp.real)  # Freq

    return np.asarray(result)

cpdef unsigned long long find_signal_start(float[::1] demod_samples, int mod_type):

    cdef unsigned long long i, ns, l
    cdef float dsample
    cdef int has_oversteuern, conseq_noise, conseq_not_noise, behind_oversteuern
    cdef float NOISE = get_noise_for_mod_type(mod_type)

    has_oversteuern = 0
    behind_oversteuern = 0
    conseq_noise = 0
    conseq_not_noise = 0

    ns = len(demod_samples)
    l = 100
    if ns < 100:
        l = ns

    for i in range(0, l):
        dsample = demod_samples[i]
        if dsample > NOISE:
            has_oversteuern = 1
            break

    for i in range(0, ns):
        dsample = demod_samples[i]

        if dsample == NOISE:
            conseq_noise += 1
            conseq_not_noise = 0
        else:
            conseq_noise = 0
            conseq_not_noise += 1

        if has_oversteuern == 1:
            if has_oversteuern and conseq_noise > 100:
                behind_oversteuern = 1

            if behind_oversteuern and conseq_not_noise == 3:
                return i - 3

        elif conseq_not_noise == 3:
            return i -3

    return 0

cpdef unsigned long long find_signal_end(float[::1] demod_samples, int mod_type):

    cdef unsigned long long i
    cdef float dsample
    cdef int conseq_not_noise = 0
    cdef float NOISE = get_noise_for_mod_type(mod_type)
    cdef unsigned long long ns = len(demod_samples)

    for i in range(ns, 0, -1):
        dsample = demod_samples[i]

        if dsample > NOISE:
            conseq_not_noise += 1

        if conseq_not_noise == 3:
            return i + 3

    return ns

cpdef unsigned long long[:, ::1] grab_pulse_lens(float[::1] samples,
                                                 float treshold, unsigned int tolerance, int mod_type):
    """
    Holt sich die Pulslängen aus den quadraturdemodulierten Samples
    @param samples: Samples nach der QAD
    @param treshold: Alles über der Treshold ist ein Einserpuls, alles darunter 0er Puls
    @return: Ein 2D Array arr.
    arr[i] gibt Position an.
    arr[i][0] gibt an ob Einspuls (arr[i][0] = 1) Nullpuls (arr[i][0] = 0) Pause (arr[i][0] = 42)
    arr[i][1] gibt die Länge des Pulses bzw. der Pause an.
    """
    cdef unsigned long long i, ns, pulselen = 0
    cdef unsigned long long cur_index = 0, conseq_ones = 0, conseq_zeros = 0, conseq_pause = 0
    cdef float s, s_prev
    cdef int cur_state
    cdef float NOISE = get_noise_for_mod_type(mod_type)
    ns = len(samples)

    cdef unsigned long long[:, ::1] result = np.empty((ns, 2), dtype=np.uint64, order="C")
    if ns == 0:
        return result

    s_prev = samples[0]
    if s_prev == NOISE:
        cur_state = 42
    elif s_prev > treshold:
        cur_state = 1
    else:
        cur_state = 0

    for i in range(ns-1):
        pulselen += 1
        s = samples[i]
        if s == NOISE:
            conseq_pause += 1
            conseq_ones = 0
            conseq_zeros = 0
            if cur_state == 42: continue
        elif s > treshold:
            conseq_ones += 1
            conseq_zeros = 0
            conseq_pause = 0
            if cur_state == 1: continue
        else:
            conseq_zeros += 1
            conseq_ones = 0
            conseq_pause = 0
            if cur_state == 0: continue

        if conseq_ones > tolerance:
            result[cur_index, 0] = cur_state
            result[cur_index, 1] = pulselen - tolerance
            cur_index += 1
            pulselen = tolerance
            cur_state = 1

        elif conseq_zeros > tolerance:
            result[cur_index, 0] = cur_state
            result[cur_index, 1] = pulselen - tolerance
            cur_index += 1
            pulselen = tolerance
            cur_state = 0

        elif conseq_pause > tolerance:
            result[cur_index, 0] = cur_state
            result[cur_index, 1] = pulselen - tolerance
            cur_index += 1
            pulselen = tolerance
            cur_state = 42

    # Letzen anfügen
    cdef unsigned long long len_result = len(result)
    if cur_index < len_result:
        result[cur_index, 0] = cur_state
        result[cur_index, 1] = pulselen
        cur_index += 1

    if cur_index > len_result:
        cur_index = len_result

    return result[:cur_index]

cdef class Symbol:
    cdef public str name
    cdef public int nbits
    cdef public int pulsetype
    cdef public unsigned long long nsamples # Num Samples for this Symbol. Needed in Modulator.
    def __init__(self, str name, int nbits, int pulsetype, unsigned long long nsamples):
        """
        :param nbits: Number of bits this Symbol covers
        :param name: Name of the symbol (one char)
        :param pulsetype: 0 für 0er Puls, 1 für 1er Puls
        :return:
        """
        self.name = name
        self.pulsetype = pulsetype
        self.nbits = nbits
        self.nsamples = nsamples

    def __repr__(self):
        return "{0} ({1}:{2})".format(self.name, self.pulsetype, self.name)

    def __deepcopy__(self, memo):
        result = Symbol(self.name, self.nbits, self.pulsetype, self.nsamples)
        memo[id(self)] = result
        return result



cpdef unsigned long long estimate_bit_len(float[::1] qad_samples, float qad_center, int tolerance, int mod_type):

    start = find_signal_start(qad_samples, mod_type)
    cdef unsigned long long[:, ::1] ppseq = grab_pulse_lens(qad_samples[start:], qad_center, tolerance, mod_type)
    cdef unsigned long long i = 0
    cdef unsigned long long l = len(ppseq)
    for i in range(0, l):
        if ppseq[i, 0] == 1:
            return ppseq[i, 1] # first pulse after pause

    return 100

cpdef int find_nearest_center(float sample, float[::1] centers, int num_centers) nogil:
    cdef int i
    cdef float center
    cdef int result = 0
    cdef float min_diff = 99999
    cdef float cur_diff = 0

    for i in range(0, num_centers):
        center = centers[i]
        cur_diff = (sample - center) * (sample - center)
        if cur_diff < min_diff:
            min_diff = cur_diff
            result = i

    return result


from libc.stdlib cimport malloc, free

cdef:
    struct Cluster:
        double sum
        unsigned long long int nitems

cpdef float estimate_qad_center(float[::1] samples, unsigned int num_centers):
    """
    Estimate the centers using Lloyds algorithm
    Use more centers for ks clipping

    :param samples:
    :param num_centers:
    :return:
    """
    cdef unsigned long long nsamples = len(samples)
    if nsamples == 0:
        return 0

    cdef Cluster *clusters = <Cluster *>malloc(num_centers * sizeof(Cluster))

    cdef unsigned long long i

    for i in range(0, num_centers):
        clusters[i].nitems = 0
        clusters[i].sum = 0

    cdef:
        tuple tmp = util.minmax(samples)
        float first_center = tmp[0]
        float last_center = tmp[1]

        float[::1] centers = np.array([ first_center+i*(last_center-first_center)/(num_centers-1)
                                       for i in range(0, num_centers) ], dtype=np.float32)
        float sample
        int center_index = 0



    for i in range(0, nsamples):
        sample = samples[i]
        center_index = find_nearest_center(sample, centers, num_centers)
        clusters[center_index].sum += sample
        clusters[center_index].nitems += 1

    cdef unsigned long long[::1] cluster_lens = np.array([clusters[i].nitems for i in range(num_centers)], dtype=np.uint64)
    cdef np.ndarray[np.int64_t, ndim=1] sorted_indexes = np.argsort(cluster_lens)
    cdef float center1, center2
    cdef int index1 = sorted_indexes[len(sorted_indexes)-1]
    cdef int index2 = sorted_indexes[len(sorted_indexes)-2]

    if clusters[index1].nitems > 0:
        center1 = clusters[index1].sum / clusters[index1].nitems # Cluster mit den meisten Einträgen
    else:
        center1 = 0

    if clusters[index2].nitems > 0:
        center2 = clusters[index2].sum / clusters[index2].nitems # Cluster mit zweitmeisten Einträgen
    else:
        center2 = 0

    free(clusters)
    return (center1 + center2)/2



from PyQt5.QtCore import QModelIndex
from PyQt5.QtGui import QContextMenuEvent, QIcon
from PyQt5.QtWidgets import QTreeView, QInputDialog, QMessageBox, QMenu


class DirectoryTreeView(QTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)

    def create_directory(self):
        index = self.model().mapToSource(self.rootIndex())  # type: QModelIndex

        if not index.isValid():
            return

        model = self.model().sourceModel()
        dir_name, ok = QInputDialog.getText(self, self.tr("Create Directory"), self.tr("Directory name"))

        if ok and len(dir_name) > 0:
            if not model.mkdir(index, dir_name).isValid():
                QMessageBox.information(self, self.tr("Create Directoy"), self.tr("Failed to create the directory"))

    def remove(self):
        index = self.model().mapToSource(self.currentIndex())  # type: QModelIndex
        if not index.isValid():
            return

        model = self.model().sourceModel()
        if model.fileInfo(index).isDir():
            ok = model.rmdir(index)
        else:
            ok = model.remove(index)

        if not ok:
            QMessageBox.information(self, self.tr("Remove"),
                                    self.tr("Failed to remove {0}".format(model.fileName(index))))

    def create_context_menu(self) -> QMenu:
        menu = QMenu(self)
        new_dir_action = menu.addAction("New Directory")
        new_dir_action.setIcon(QIcon.fromTheme("folder"))
        new_dir_action.triggered.connect(self.create_directory)

        del_action = menu.addAction("Delete")
        del_action.setIcon(QIcon.fromTheme("edit-delete"))
        del_action.triggered.connect(self.remove)

        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        menu = self.create_context_menu()
        menu.exec_(self.mapToGlobal(event.pos()))



import random
import string
from copy import deepcopy

from urh import constants
from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.signalprocessing.Ruleset import Ruleset
from urh.util.Logger import logger
import xml.etree.ElementTree as ET


class MessageType(list):
    """
    A message type is a list of protocol fields.

    """

    __slots__ = ["name", "__id", "assigned_by_ruleset", "ruleset", "assigned_by_logic_analyzer"]

    def __init__(self, name: str, iterable=None, id=None, ruleset=None):
        iterable = iterable if iterable else []
        super().__init__(iterable)

        self.name = name
        self.__id = ''.join(
            random.choice(string.ascii_letters + string.digits) for _ in range(50)) if id is None else id

        self.assigned_by_logic_analyzer = False
        self.assigned_by_ruleset = False
        self.ruleset = Ruleset() if ruleset is None else ruleset

    def __hash__(self):
        return hash(super)

    def __getitem__(self, index) -> ProtocolLabel:
        return super().__getitem__(index)

    def __repr__(self):
        return self.name + " " + super().__repr__()

    def __eq__(self, other):
        if isinstance(other, MessageType):
            return self.id == other.id
        else:
            return super().__eq__(other)

    @property
    def assign_manually(self):
        return not self.assigned_by_ruleset

    @property
    def id(self) -> str:
        return self.__id

    @property
    def unlabeled_ranges(self):
        """

        :rtype: list[(int,int)]
        """
        return self.__get_unlabeled_ranges_from_labels(self)

    @staticmethod
    def __get_unlabeled_ranges_from_labels(labels):
        """

        :type labels: list of ProtocolLabel
        :rtype: list[(int,int)]
        """
        start = 0
        result = []
        for lbl in labels:
            if lbl.start > start:
                result.append((start, lbl.start))
            start = lbl.end
        result.append((start, None))
        return result

    def unlabeled_ranges_with_other_mt(self, other_message_type):
        """

        :type other_message_type: MessageType
        :rtype: list[(int,int)]
        """
        labels = self + other_message_type
        labels.sort()
        return self.__get_unlabeled_ranges_from_labels(labels)

    def append(self, lbl: ProtocolLabel):
        super().append(lbl)
        self.sort()

    def add_protocol_label(self, start: int, end: int, name=None, color_ind=None,
                           auto_created=False, type: FieldType = None) -> ProtocolLabel:

        name = "" if not name else name
        used_colors = [p.color_index for p in self]
        avail_colors = [i for i, _ in enumerate(constants.LABEL_COLORS) if i not in used_colors]

        if color_ind is None:
            if len(avail_colors) > 0:
                color_ind = avail_colors[0]
            else:
                color_ind = random.randint(0, len(constants.LABEL_COLORS) - 1)

        proto_label = ProtocolLabel(name=name, start=start, end=end, color_index=color_ind,
                                    auto_created=auto_created, type=type)

        if proto_label not in self:
            self.append(proto_label)
            self.sort()

        return proto_label  # Return label to set editor focus after adding

    def add_label(self, lbl: ProtocolLabel, allow_overlapping=True):
        if allow_overlapping or not any(lbl.overlaps_with(l) for l in self):
            self.add_protocol_label(lbl.start, lbl.end, name=lbl.name, color_ind=lbl.color_index)

    def remove(self, lbl: ProtocolLabel):
        if lbl in self:
            super().remove(lbl)
        else:
            logger.warning(lbl.name + " is not in set, so cant be removed")

    def to_xml(self) -> ET.Element:
        result = ET.Element("message_type", attrib={"name": self.name, "id": self.id,
                                                    "assigned_by_ruleset": "1" if self.assigned_by_ruleset else "0",
                                                    "assigned_by_logic_analyzer": "1" if self.assigned_by_logic_analyzer else "0"})
        for lbl in self:
            result.append(lbl.to_xml(-1))

        result.append(self.ruleset.to_xml())

        return result

    @staticmethod
    def from_xml(tag: ET.Element):
        field_types_by_type_id = {ft.id: ft for ft in FieldType.load_from_xml()}

        name = tag.get("name", "blank")
        id = tag.get("id", None)
        assigned_by_ruleset = bool(int(tag.get("assigned_by_ruleset", 0)))
        assigned_by_logic_analyzer = bool(int(tag.get("assigned_by_logic_analyzer", 0)))
        labels = []
        for lbl_tag in tag.findall("label"):
            labels.append(ProtocolLabel.from_xml(lbl_tag, field_types_by_type_id=field_types_by_type_id))
        result = MessageType(name=name, iterable=labels, id=id, ruleset=Ruleset.from_xml(tag.find("ruleset")))
        result.assigned_by_ruleset = assigned_by_ruleset
        result.assigned_by_logic_analyzer = assigned_by_logic_analyzer

        return result

    def copy_for_fuzzing(self):
        result = deepcopy(self)
        for lbl in result:
            lbl.fuzz_values = []
            lbl.fuzz_created = True
        return result



import os
import time

import numpy as np
from PyQt5.QtCore import pyqtSlot, pyqtSignal, QObject
from PyQt5.QtWidgets import QApplication

from urh.cythonext.signalFunctions import grab_pulse_lens
from urh.dev.BackendHandler import BackendHandler, Backends
from urh.dev.VirtualDevice import VirtualDevice, Mode
from urh.signalprocessing.Message import Message
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Signal import Signal


class ProtocolSniffer(ProtocolAnalyzer, QObject):
    """
    This class is used for live sniffing a protocol
    with certain signal parameters.
    """
    started = pyqtSignal()
    stopped = pyqtSignal()

    def __init__(self, bit_len: int, center: float, noise: float, tolerance: int,
                 modulation_type: int, device: str, testing_mode=False):
        signal = Signal("", "LiveSignal")
        signal.bit_len = bit_len
        signal.qad_center = center
        signal.noise_threshold = noise
        signal.tolerance = tolerance
        signal.silent_set_modulation_type(modulation_type)
        ProtocolAnalyzer.__init__(self, signal)
        QObject.__init__(self, None)

        self.backend_handler = BackendHandler(testing_mode=testing_mode)
        self.rcv_device = VirtualDevice(self.backend_handler, device, Mode.receive,
                                        is_ringbuffer=False, raw_mode=False)

        self.rcv_device.index_changed.connect(self.on_rcv_thread_index_changed)
        self.rcv_device.started.connect(self.__emit_started)
        self.rcv_device.stopped.connect(self.__emit_stopped)

        self.rel_symbol_len = self._read_symbol_len()

        self.data_cache = []
        self.conseq_non_data = 0
        self.reading_data = False

        self.store_messages = True

        self.__sniff_file = ""
        self.__store_data = True

    def plain_to_string(self, view: int, show_pauses=True, start=0) -> str:
        """

        :param start: First message to begin with
        :param show_pauses: Show pauses in output?
        :param view: 0 - Bits ## 1 - Hex ## 2 - ASCII
        """
        return '\n'.join(msg.view_to_string(view, False, show_pauses) for
                         msg in self.messages[start:])

    @property
    def sniff_file(self):
        return self.__sniff_file

    @sniff_file.setter
    def sniff_file(self, val):
        self.__sniff_file = val
        if self.__sniff_file:
            self.__store_data = False

    @property
    def device_name(self):
        return self.rcv_device.name

    @device_name.setter
    def device_name(self, value: str):
        if value != self.rcv_device.name:
            self.rcv_device.free_data()
            self.rcv_device = VirtualDevice(self.backend_handler, value, Mode.receive,
                                            device_ip="192.168.10.2", is_ringbuffer=False, raw_mode=False)
            self.rcv_device.index_changed.connect(self.on_rcv_thread_index_changed)
            self.rcv_device.started.connect(self.__emit_started)
            self.rcv_device.stopped.connect(self.__emit_stopped)

    def sniff(self):
        self.rcv_device.start()

    @pyqtSlot(int, int)
    def on_rcv_thread_index_changed(self, old_index, new_index):
        old_nmsgs = len(self.messages)
        if self.rcv_device.backend in (Backends.native, Backends.grc):
            if old_index == new_index:
                return
            self.__demodulate_data(self.rcv_device.data[old_index:new_index])
        elif self.rcv_device.backend == Backends.network:
            # We receive the bits here
            for bit_str in self.rcv_device.data:
                self.messages.append(Message.from_plain_bits_str(bit_str, {}))

            self.rcv_device.free_data()  # do not store received bits twice

        self.qt_signals.data_sniffed.emit(old_nmsgs)

        if self.sniff_file and not os.path.isdir(self.sniff_file):
            with open(self.sniff_file, "a") as myfile:
                myfile.write("\n".join(self.plain_bits_str))

        if not self.__store_data:
            self.messages[:] = []

    def __demodulate_data(self, data):
        """
        Demodulates received IQ data and adds demodulated bits to messages
        :param data:
        :return:
        """
        signal = self.signal

        if self.__are_bits_in_data(data):
            self.reading_data = True
        elif self.conseq_non_data == 5:
            self.reading_data = False
            self.conseq_non_data = 0
        else:
            self.conseq_non_data += 1

        if self.reading_data:
            self.data_cache.append(data)
            return
        elif len(self.data_cache) == 0:
            return

        signal._fulldata = np.concatenate(self.data_cache)
        del self.data_cache[:]
        signal._qad = None

        bit_len = signal.bit_len
        ppseq = grab_pulse_lens(signal.qad, signal.qad_center,
                                signal.tolerance, signal.modulation_type)

        bit_data, pauses, bit_sample_pos = self._ppseq_to_bits(ppseq, bit_len, self.rel_symbol_len)

        i = 0
        first_msg = True

        for bits, pause in zip(bit_data, pauses):
            if first_msg or self.messages[-1].pause > 8 * bit_len:
                # Create new Message
                middle_bit_pos = bit_sample_pos[i][int(len(bits) / 2)]
                start, end = middle_bit_pos, middle_bit_pos + bit_len
                rssi = np.mean(np.abs(signal._fulldata[start:end]))
                message = Message(bits, pause, bit_len=bit_len, rssi=rssi, message_type=self.default_message_type)
                self.messages.append(message)
                first_msg = False
            else:
                # Append to last message
                message = self.messages[-1]
                nzeros = int(np.round(message.pause / bit_len))
                message.plain_bits.extend([False] * nzeros)
                message.plain_bits.extend(bits)
                message.pause = pause
            i += 1

    def __are_bits_in_data(self, data):
        signal = self.signal
        signal._fulldata = data
        signal._qad = None

        bit_len = signal.bit_len
        ppseq = grab_pulse_lens(signal.qad, signal.qad_center,
                                signal.tolerance, signal.modulation_type)

        bit_data, pauses, _ = self._ppseq_to_bits(ppseq, bit_len, self.rel_symbol_len)

        return bool(bit_data)

    def stop(self):
        self.rcv_device.stop("Stopping receiving due to user interaction")
        QApplication.processEvents()
        time.sleep(0.1)

    def clear(self):
        del self.data_cache[:]
        del self.messages[:]

    def __emit_started(self):
        self.started.emit()

    def __emit_stopped(self):
        self.stopped.emit()



from PyQt5.QtCore import Qt, QPointF
from PyQt5.QtGui import QColor, QPen
from PyQt5.QtWidgets import QGraphicsRectItem

from urh import constants


class ROI(QGraphicsRectItem):
    def __init__(self, *args, fillcolor, opacity, parent=None):
        self.padding = 0
        if len(args) == 0:
            super().__init__(parent)
        elif len(args) == 1:
            super().__init__(args[0], parent)
        elif len(args) == 4:
            x0 = args[0]
            y0 = args[1]
            w = args[2]
            h = args[3]
            ymin = y0 + self.padding * h
            roi_height = (1 - 2 * self.padding) * h
            super().__init__(x0, ymin, w, roi_height, parent)

        self.finished = False
        self.selected_edge = None
        """:type: int """
        self.resizing = False

        self.setBrush(fillcolor)
        self.setPen(QPen(QColor(Qt.transparent), Qt.FlatCap))
        self.setOpacity(opacity)

    @property
    def is_empty(self) -> bool:
        return self.width == 0

    @property
    def width(self):
        return self.rect().width()

    @width.setter
    def width(self, value):
        if value == self.width:
            return

        r = self.rect()
        r.setWidth(value)
        self.setRect(r)

    @property
    def height(self):
        return self.rect().height()

    @height.setter
    def height(self, value):
        if value == self.height:
            return

        r = self.rect()
        r.setHeight(value)
        self.setRect(r)

    @property
    def x(self):
        return self.rect().x()

    def setX(self, p_float):
        if p_float == self.x:
            return

        r = self.rect()
        r.setX(p_float)
        self.setRect(r)

    @property
    def start(self):
        if self.width < 0:
            return self.x + self.width
        else:
            return self.x

    @start.setter
    def start(self, value):
        self.setX(value)

    @property
    def end(self):
        if self.width < 0:
            return self.x
        else:
            return self.x + self.width

    @end.setter
    def end(self, value):
        self.width = value - self.start

    @property
    def y(self):
        return self.rect().y()

    def setY(self, p_float):
        if p_float == self.y:
            return

        r = self.rect()
        r.setY(p_float)
        self.setRect(r)

    def is_in_roi(self, pos: QPointF):
        x1 = self.rect().x()
        x2 = x1 + self.rect().width()
        y1 = self.rect().y()
        y2 = y1 + self.rect().width()

        if x1 < pos.x() < x2 and y1 < pos.y() < y2:
            return True

        return False

    def get_selected_edge(self, pos: QPointF, width_view: float):
        """
        Bestimmt auf welcher Ecke der ROI der Mauszeiger gerade ist.
        0 = links, 1 = rechts

        :param pos: In die Szene gemappte Position des Mauszeigers
        """
        x1 = self.rect().x()
        x2 = x1 + self.rect().width()
        y1 = self.rect().y()
        y2 = y1 + self.rect().height()
        x = pos.x()
        y = pos.y()

        if x1 - 0.025 * width_view < x < x1 + 0.025 * width_view and y1 < y < y2:
            self.selected_edge = 0
            return 0

        if x2 - 0.025 * width_view < x < x2 + 0.025 * width_view and y1 < y < y2:
            self.selected_edge = 1
            return 1

        self.selected_edge = None
        return None

    def clear(self):
        self.width = 0
        self.resizing = False
        self.selected_edge = None
        self.finished = False






import socket
import numpy as np
import time

import zmq

from urh.dev.gr.AbstractBaseThread import AbstractBaseThread
from urh.util.Logger import logger


class SpectrumThread(AbstractBaseThread):
    def __init__(self, freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, ip='127.0.0.1', parent=None):
        super().__init__(freq, sample_rate, bandwidth, gain, if_gain, baseband_gain, True, ip, parent)
        self.buf_size = 10**5
        self.data = np.zeros(self.buf_size, dtype=np.complex64)
        self.x = None
        self.y = None

    def run(self):
        logger.debug("Spectrum Thread: Init Process")
        self.initialize_process()
        logger.debug("Spectrum Thread: Process Intialized")
        self.init_recv_socket()
        logger.debug("Spectrum Thread: Socket initialized")

        recv = self.socket.recv
        rcvd = b""

        try:
            logger.debug("Spectrum Thread: Enter main loop")
            while not self.isInterruptionRequested():
                try:
                    rcvd += recv(32768)  # Receive Buffer = 32768 Byte
                except (zmq.error.ContextTerminated, ConnectionResetError):
                    self.stop("Stopped receiving, because connection was reset")
                    return
                except OSError as e:  # https://github.com/jopohl/urh/issues/131
                    logger.warning("Error occurred", str(e))

                if len(rcvd) < 8:
                    self.stop("Stopped receiving, because no data transmitted anymore")
                    return

                if len(rcvd) % 8 != 0:
                    continue

                try:
                    tmp = np.fromstring(rcvd, dtype=np.complex64)

                    len_tmp = len(tmp)

                    if self.data is None:
                        self.data = np.zeros(self.buf_size, dtype=np.complex64)  # type: np.ndarray

                    if self.current_index + len_tmp >= len(self.data):
                        self.data[self.current_index:] = tmp[:len(self.data) - self.current_index]
                        tmp = tmp[len(self.data) - self.current_index:]
                        w = np.abs(np.fft.fft(self.data))
                        freqs = np.fft.fftfreq(len(w), 1 / self.sample_rate)
                        idx = np.argsort(freqs)
                        self.x = freqs[idx].astype(np.float32)
                        self.y = w[idx].astype(np.float32)

                        self.data = np.zeros(len(self.data), dtype=np.complex64)
                        self.data[0:len(tmp)] = tmp
                        self.current_index = len(tmp)
                        continue

                    self.data[self.current_index:self.current_index + len_tmp] = tmp
                    self.current_index += len_tmp
                    rcvd = b""
                except ValueError:
                    self.stop("Could not receive data. Is your Hardware ok?")
        except RuntimeError as e:
            logger.error("Spectrum thread crashed", str(e.args))



import copy
import itertools
import xml
from xml.dom import minidom
import xml.etree.ElementTree as ET

import numpy

from urh.cythonext.signalFunctions import Symbol
from urh.models.ProtocolTreeItem import ProtocolTreeItem
from urh.signalprocessing.Modulator import Modulator
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Message import Message
from urh.signalprocessing.ProtocolGroup import ProtocolGroup
from urh.signalprocessing.encoder import Encoder
from urh.util.Formatter import Formatter
from urh.util.Logger import logger
from enum import Enum

class FuzzMode(Enum):
    successive = 0
    concurrent = 1
    exhaustive = 2

class ProtocolAnalyzerContainer(ProtocolAnalyzer):
    """
    A container to manage several ProtocolAnalyzers.
    This class is used by the generator to manage distinct protocols
    """

    def __init__(self, modulators):
        """

        :type modulators: list of Modulator
        """
        super().__init__(None)
        self.modulators = modulators

        self.fuzz_pause = 10000

        self.__group = ProtocolGroup("GeneratorGroup")
        self.__group.add_protocol_item(ProtocolTreeItem(self, None)) # Warning: parent is None

    @property
    def protocol_labels(self):
        result = list(set(lbl for msg in self.messages for lbl in msg.message_type))
        result.sort()
        return result

    @property
    def multiple_fuzz_labels_per_message(self):
        return any(len(msg.active_fuzzing_labels) > 1 for msg in self.messages)

    def insert_protocol_analyzer(self, index: int, proto_analyzer: ProtocolAnalyzer):

        messages = [Message(plain_bits=copy.copy(msg.decoded_bits), pause=msg.pause,
                            message_type=copy.deepcopy(msg.message_type),
                            rssi=msg.rssi, modulator_indx=0, decoder=msg.decoder, bit_len=msg.bit_len, participant=msg.participant)
                  for msg in proto_analyzer.messages if msg]

        self.messages[index:0] = messages
        self.used_symbols |= proto_analyzer.used_symbols

        if len(self.pauses) > 0:
            self.fuzz_pause = self.pauses[0]

    def duplicate_line(self, row: int):
        try:
            self.messages.insert(row + 1, copy.deepcopy(self.messages[row]))
            self.qt_signals.line_duplicated.emit()
        except Exception as e:
            logger.error("Duplicating line ", str(e))

    def fuzz(self, mode: FuzzMode, default_pause=None):
        result = []
        appd_result = result.append

        for i, msg in enumerate(self.messages):
            labels = msg.active_fuzzing_labels
            appd_result(msg)

            if mode == FuzzMode.successive:
                combinations = [[(l.start, l.end, fuzz_val)] for l in labels for fuzz_val in l.fuzz_values[1:]]
            elif mode == FuzzMode.concurrent:
                nval = numpy.max([len(l.fuzz_values) for l in labels]) if labels else 0
                f = lambda index, label: index if index < len(label.fuzz_values) else 0
                combinations = [[(l.start, l.end, l.fuzz_values[f(j, l)]) for l in labels] for j in range(1, nval)]
            elif mode == FuzzMode.exhaustive:
                pool = [[(l.start, l.end, fv) for fv in l.fuzz_values[1:]] for l in labels]
                combinations = itertools.product(*pool) if labels else []
            else:
                raise ValueError("Unknown fuzz mode")

            for combination in combinations:
                cpy_bits = msg.plain_bits[:]
                for start, end, fuz_val in combination:
                    cpy_bits[start:end] = [True if bit == "1" else False for bit in fuz_val]

                pause = default_pause if default_pause is not None else msg.pause
                fuz_msg = Message(plain_bits=cpy_bits, pause=pause,
                                  rssi=msg.rssi, message_type=msg.message_type.copy_for_fuzzing(),
                                  modulator_indx=msg.modulator_indx,
                                  decoder=msg.decoder, fuzz_created=True)
                appd_result(fuz_msg)

        self.messages = result
        """:type: list of Message """

    def fuzz_successive(self, default_pause=None):
        """
        Führt ein sukzessives Fuzzing über alle aktiven Fuzzing Label durch.
        Sequentiell heißt, ein Label wird durchgefuzzt und alle anderen Labels bleiben auf Standardwert.
        Das entspricht dem Vorgang nacheinander immer nur ein Label aktiv zu setzen.
        """
        self.fuzz(FuzzMode.successive, default_pause=default_pause)

    def fuzz_concurrent(self, default_pause=None):
        """
        Führt ein gleichzeitiges Fuzzing durch, das heißt bei mehreren Labels pro Message werden alle Labels
        gleichzeitig iteriert. Wenn ein Label keine FuzzValues mehr übrig hat,
        wird der erste Fuzzing Value (per Definition der Standardwert) genommen.
        """
        self.fuzz(FuzzMode.concurrent, default_pause=default_pause)

    def fuzz_exhaustive(self, default_pause=None):
        """
        Führt ein vollständiges Fuzzing durch. D.h. wenn es mehrere Label pro Message gibt, werden alle
        möglichen Kombinationen erzeugt (Kreuzprodukt!)
        """
        self.fuzz(FuzzMode.exhaustive, default_pause=default_pause)

    def create_fuzzing_label(self, start, end, msg_index) -> ProtocolLabel:
        fuz_lbl = self.messages[msg_index].message_type.add_protocol_label(start=start, end=end)
        return fuz_lbl

    def set_decoder_for_messages(self, decoder, messages=None):
        raise NotImplementedError("Encoding cant be set in Generator!")

    def to_xml_file(self, filename: str, decoders=None, participants=None, tag_name="fuzz_profile", include_message_types=True, write_bits=True):
        super().to_xml_file(filename=filename, decoders=None, participants=participants, tag_name=tag_name, include_message_types=include_message_types, write_bits=write_bits)

    def from_xml_file(self, filename: str, read_bits=True):
        super().from_xml_file(filename=filename, read_bits=read_bits)

    def clear(self):
        self.messages[:] = []


import numpy as np

from urh.dev.native.Device import Device

try:
    from urh.dev.native.lib import rtlsdr
except ImportError:
    import urh.dev.native.lib.rtlsdr_fallback as rtlsdr
from urh.util.Logger import logger

class RTLSDR(Device):
    BYTES_PER_SAMPLE = 2  # RTLSDR device produces 8 bit unsigned IQ data

    @staticmethod
    def receive_sync(data_connection, ctrl_connection, device_number: int, center_freq: int, sample_rate: int,
                     gain: int, freq_correction: int, direct_sampling_mode: int):
        ret = rtlsdr.open(device_number)
        ctrl_connection.send("open:" + str(ret))

        ret = rtlsdr.set_center_freq(center_freq)
        ctrl_connection.send("set_center_freq:" + str(ret))

        ret = rtlsdr.set_sample_rate(sample_rate)
        ctrl_connection.send("set_sample_rate:" + str(ret))

        ret = rtlsdr.set_tuner_gain(10*gain)
        ctrl_connection.send("set_tuner_gain to {0}:{1}".format(gain, ret))

        ret = rtlsdr.set_freq_correction(freq_correction)
        ctrl_connection.send("set_freq_correction to {0}:{1}".format(freq_correction, ret))

        ret = rtlsdr.set_direct_sampling(direct_sampling_mode)
        ctrl_connection.send("set_direct_sampling_mode to {0}:{1}".format(direct_sampling_mode, ret))

        ret = rtlsdr.reset_buffer()
        ctrl_connection.send("reset_buffer:" + str(ret))

        exit_requested = False

        while not exit_requested:
            while ctrl_connection.poll():
                result = RTLSDR.process_command(ctrl_connection.recv())
                if result == "stop":
                    exit_requested = True
                    break

            if not exit_requested:
                data_connection.send_bytes(rtlsdr.read_sync())

        logger.debug("RTLSDR: closing device")
        ret = rtlsdr.close()
        ctrl_connection.send("close:" + str(ret))

        data_connection.close()
        ctrl_connection.close()

    @staticmethod
    def process_command(command):
        logger.debug("RTLSDR: {}".format(command))
        if command == "stop":
            return "stop"

        tag, value = command.split(":")
        if tag == "center_freq":
            logger.info("RTLSDR: Set center freq to {0}".format(int(value)))
            return rtlsdr.set_center_freq(int(value))

        elif tag == "rf_gain":
            logger.info("RTLSDR: Set tuner gain to {0}".format(int(value)))
            return rtlsdr.set_tuner_gain(10*int(value))  # calculate *10 for API

        elif tag == "sample_rate":
            logger.info("RTLSDR: Set sample_rate to {0}".format(int(value)))
            return rtlsdr.set_sample_rate(int(value))

        elif tag == "tuner_bandwidth":
            logger.info("RTLSDR: Set bandwidth to {0}".format(int(value)))
            return rtlsdr.set_tuner_bandwidth(int(value))

        elif tag == "freq_correction":
            logger.info("RTLSDR: Set freq_correction to {0}".format(int(value)))
            return rtlsdr.set_freq_correction(int(value))

        elif tag == "direct_sampling_mode":
            logger.info("RTLSDR: Set direct_sampling_mode to {0}".format(int(value)))
            return rtlsdr.set_direct_sampling(int(value))

    def __init__(self, freq, gain, srate, device_number, is_ringbuffer=False):
        super().__init__(center_freq=freq, sample_rate=srate, bandwidth=0,
                         gain=gain, if_gain=1, baseband_gain=1, is_ringbuffer=is_ringbuffer)

        self.success = 0

        self.receive_process_function = RTLSDR.receive_sync

        self.bandwidth_is_adjustable = hasattr(rtlsdr,
                                               "set_tuner_bandwidth")  # e.g. not in Manjaro Linux / Ubuntu 14.04

        self.device_number = device_number

    @property
    def receive_process_arguments(self):
        return self.child_data_conn, self.child_ctrl_conn, self.device_number, self.frequency, self.sample_rate, \
               self.gain, self.freq_correction, self.direct_sampling_mode

    def set_device_bandwidth(self, bandwidth):
        if hasattr(rtlsdr, "set_tuner_bandwidth"):
            self.parent_ctrl_conn.send("tuner_bandwidth:{}".format(int(bandwidth)))
        else:
            logger.warning("Setting the bandwidth is not supported by your RTL-SDR driver version.")

    @staticmethod
    def unpack_complex(buffer, nvalues: int):
        """
        The raw, captured IQ data is 8 bit unsigned data.

        :return:
        """
        result = np.empty(nvalues, dtype=np.complex64)
        unpacked = np.frombuffer(buffer, dtype=[('r', np.uint8), ('i', np.uint8)])
        result.real = (unpacked['r'] / 127.5) - 1.0
        result.imag = (unpacked['i'] / 127.5) - 1.0
        return result

    @staticmethod
    def pack_complex(complex_samples: np.ndarray):
        return (127.5 * (complex_samples.view(np.float32) + 1.0)).astype(np.uint8).tostring()



import numpy as np
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QMessageBox

from urh.LiveSceneManager import LiveSceneManager
from urh.controller.SendRecvDialogController import SendRecvDialogController
from urh.dev.VirtualDevice import Mode, VirtualDevice
from urh.util import FileOperator
from urh.util.Formatter import Formatter


class ReceiveDialogController(SendRecvDialogController):
    files_recorded = pyqtSignal(list)

    def __init__(self, project_manager, parent=None, testing_mode=False):
        try:
            super().__init__(project_manager, is_tx=False, parent=parent, testing_mode=testing_mode)
        except ValueError:
            return

        self.update_interval = 25

        self.graphics_view = self.ui.graphicsViewReceive
        self.ui.stackedWidget.setCurrentIndex(0)
        self.hide_send_ui_items()
        self.already_saved = True
        self.recorded_files = []

        # set really in on_device_started
        self.scene_manager = None  # type: LiveSceneManager

        self.init_device()
        self.set_bandwidth_status()

        self.graphics_view.setScene(self.scene_manager.scene)
        self.graphics_view.scene_manager = self.scene_manager

        self.create_connects()

    def create_connects(self):
        super().create_connects()
        self.ui.btnSave.clicked.connect(self.on_save_clicked)

    def save_before_close(self):
        if not self.already_saved and self.device.current_index > 0:
            reply = QMessageBox.question(self, self.tr("Save data?"),
                                         self.tr("Do you want to save the data you have captured so far?"),
                                         QMessageBox.Yes | QMessageBox.No | QMessageBox.Abort)
            if reply == QMessageBox.Yes:
                self.on_save_clicked()
            elif reply == QMessageBox.Abort:
                return False

        self.files_recorded.emit(self.recorded_files)
        return True

    def update_view(self):
        if super().update_view():
            self.scene_manager.end = self.device.current_index
            self.scene_manager.init_scene()
            self.scene_manager.show_full_scene()
            self.graphics_view.update()

    def init_device(self):
        device_name = self.ui.cbDevice.currentText()
        self.device = VirtualDevice(self.backend_handler, device_name, Mode.receive,
                                    device_ip="192.168.10.2", parent=self)
        self._create_device_connects()

        self.scene_manager = LiveSceneManager(np.array([]), parent=self)

    @pyqtSlot()
    def on_start_clicked(self):
        super().on_start_clicked()
        self.device.start()

    @pyqtSlot()
    def on_device_started(self):
        self.scene_manager.plot_data = self.device.data.real if self.device.data is not None else None

        super().on_device_started()

        self.already_saved = False
        self.ui.btnStart.setEnabled(False)
        self.set_device_ui_items_enabled(False)

    @pyqtSlot()
    def on_clear_clicked(self):
        self.scene_manager.clear_path()
        self.reset()

    @pyqtSlot()
    def on_save_clicked(self):
        data = self.device.data[:self.device.current_index]

        dev = self.device
        big_val = Formatter.big_value_with_suffix
        initial_name = "{0} {1}Hz {2}Sps {3}Hz.complex".format(dev.name, big_val(dev.frequency),
                                                               big_val(dev.sample_rate),
                                                               big_val(dev.bandwidth)).replace(
            Formatter.local_decimal_seperator(), "_").replace("_000", "")

        filename = FileOperator.save_data_dialog(initial_name, data, parent=self)
        self.already_saved = True
        if filename is not None and filename not in self.recorded_files:
            self.recorded_files.append(filename)



# -*- coding: utf-8 -*-

#
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Decoder(object):
    def setupUi(self, Decoder):
        Decoder.setObjectName("Decoder")
        Decoder.setWindowModality(QtCore.Qt.WindowModal)
        Decoder.resize(923, 685)
        Decoder.setModal(True)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(Decoder)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.combobox_decodings = QtWidgets.QComboBox(Decoder)
        self.combobox_decodings.setObjectName("combobox_decodings")
        self.combobox_decodings.addItem("")
        self.combobox_decodings.addItem("")
        self.horizontalLayout_2.addWidget(self.combobox_decodings)
        self.delete_decoding = QtWidgets.QPushButton(Decoder)
        self.delete_decoding.setObjectName("delete_decoding")
        self.horizontalLayout_2.addWidget(self.delete_decoding)
        self.saveas = QtWidgets.QPushButton(Decoder)
        self.saveas.setObjectName("saveas")
        self.horizontalLayout_2.addWidget(self.saveas)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout_4.addLayout(self.horizontalLayout_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_8 = QtWidgets.QLabel(Decoder)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_2.addWidget(self.label_8)
        self.basefunctions = QtWidgets.QListWidget(Decoder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.basefunctions.sizePolicy().hasHeightForWidth())
        self.basefunctions.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.basefunctions.setFont(font)
        self.basefunctions.setDragEnabled(True)
        self.basefunctions.setDragDropMode(QtWidgets.QAbstractItemView.DragOnly)
        self.basefunctions.setObjectName("basefunctions")
        self.verticalLayout_2.addWidget(self.basefunctions)
        self.label_9 = QtWidgets.QLabel(Decoder)
        self.label_9.setObjectName("label_9")
        self.verticalLayout_2.addWidget(self.label_9)
        self.additionalfunctions = QtWidgets.QListWidget(Decoder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.additionalfunctions.sizePolicy().hasHeightForWidth())
        self.additionalfunctions.setSizePolicy(sizePolicy)
        self.additionalfunctions.setDragEnabled(True)
        self.additionalfunctions.setDragDropMode(QtWidgets.QAbstractItemView.DragOnly)
        self.additionalfunctions.setObjectName("additionalfunctions")
        self.verticalLayout_2.addWidget(self.additionalfunctions)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Decoder)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.label_4 = QtWidgets.QLabel(Decoder)
        font = QtGui.QFont()
        font.setItalic(True)
        font.setStrikeOut(False)
        self.label_4.setFont(font)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.verticalLayout.addWidget(self.label_4)
        self.decoderchain = ListWidget(Decoder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.decoderchain.sizePolicy().hasHeightForWidth())
        self.decoderchain.setSizePolicy(sizePolicy)
        self.decoderchain.setAcceptDrops(True)
        self.decoderchain.setDragEnabled(True)
        self.decoderchain.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
        self.decoderchain.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.decoderchain.setTextElideMode(QtCore.Qt.ElideMiddle)
        self.decoderchain.setResizeMode(QtWidgets.QListView.Fixed)
        self.decoderchain.setViewMode(QtWidgets.QListView.ListMode)
        self.decoderchain.setObjectName("decoderchain")
        self.verticalLayout.addWidget(self.decoderchain)
        self.label_7 = QtWidgets.QLabel(Decoder)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_7.setFont(font)
        self.label_7.setAlignment(QtCore.Qt.AlignCenter)
        self.label_7.setObjectName("label_7")
        self.verticalLayout.addWidget(self.label_7)
        self.horizontalLayout.addLayout(self.verticalLayout)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.gb_infoandoptions = QtWidgets.QGroupBox(Decoder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.gb_infoandoptions.sizePolicy().hasHeightForWidth())
        self.gb_infoandoptions.setSizePolicy(sizePolicy)
        self.gb_infoandoptions.setObjectName("gb_infoandoptions")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.gb_infoandoptions)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.info = QtWidgets.QLabel(self.gb_infoandoptions)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.info.sizePolicy().hasHeightForWidth())
        self.info.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setItalic(True)
        self.info.setFont(font)
        self.info.setTextFormat(QtCore.Qt.PlainText)
        self.info.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.info.setWordWrap(True)
        self.info.setObjectName("info")
        self.verticalLayout_5.addWidget(self.info)
        self.optionWidget = QtWidgets.QStackedWidget(self.gb_infoandoptions)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.optionWidget.sizePolicy().hasHeightForWidth())
        self.optionWidget.setSizePolicy(sizePolicy)
        self.optionWidget.setObjectName("optionWidget")
        self.page_empty = QtWidgets.QWidget()
        self.page_empty.setObjectName("page_empty")
        self.optionWidget.addWidget(self.page_empty)
        self.page_redundancy = QtWidgets.QWidget()
        self.page_redundancy.setObjectName("page_redundancy")
        self.multiple = QtWidgets.QSpinBox(self.page_redundancy)
        self.multiple.setGeometry(QtCore.QRect(0, 0, 56, 23))
        self.multiple.setMinimum(2)
        self.multiple.setObjectName("multiple")
        self.label_5 = QtWidgets.QLabel(self.page_redundancy)
        self.label_5.setGeometry(QtCore.QRect(60, 0, 171, 21))
        self.label_5.setObjectName("label_5")
        self.optionWidget.addWidget(self.page_redundancy)
        self.page_carrier = QtWidgets.QWidget()
        self.page_carrier.setObjectName("page_carrier")
        self.carrier = QtWidgets.QLineEdit(self.page_carrier)
        self.carrier.setGeometry(QtCore.QRect(0, 0, 113, 23))
        self.carrier.setObjectName("carrier")
        self.label_6 = QtWidgets.QLabel(self.page_carrier)
        self.label_6.setGeometry(QtCore.QRect(120, 0, 171, 21))
        self.label_6.setObjectName("label_6")
        self.optionWidget.addWidget(self.page_carrier)
        self.page_substitution = QtWidgets.QWidget()
        self.page_substitution.setObjectName("page_substitution")
        self.gridLayout = QtWidgets.QGridLayout(self.page_substitution)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.substitution_rows = QtWidgets.QSpinBox(self.page_substitution)
        self.substitution_rows.setMinimum(1)
        self.substitution_rows.setMaximum(1000)
        self.substitution_rows.setProperty("value", 4)
        self.substitution_rows.setObjectName("substitution_rows")
        self.gridLayout.addWidget(self.substitution_rows, 0, 0, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.page_substitution)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 0, 1, 1, 1)
        self.substitution = QtWidgets.QTableWidget(self.page_substitution)
        self.substitution.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.substitution.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.substitution.setObjectName("substitution")
        self.substitution.setColumnCount(0)
        self.substitution.setRowCount(0)
        self.gridLayout.addWidget(self.substitution, 1, 0, 1, 2)
        self.optionWidget.addWidget(self.page_substitution)
        self.page_external = QtWidgets.QWidget()
        self.page_external.setObjectName("page_external")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.page_external)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_11 = QtWidgets.QLabel(self.page_external)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_3.addWidget(self.label_11)
        self.external_decoder = QtWidgets.QLineEdit(self.page_external)
        self.external_decoder.setObjectName("external_decoder")
        self.horizontalLayout_3.addWidget(self.external_decoder)
        self.btnChooseDecoder = QtWidgets.QToolButton(self.page_external)
        self.btnChooseDecoder.setObjectName("btnChooseDecoder")
        self.horizontalLayout_3.addWidget(self.btnChooseDecoder)
        self.verticalLayout_6.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_12 = QtWidgets.QLabel(self.page_external)
        self.label_12.setObjectName("label_12")
        self.horizontalLayout_4.addWidget(self.label_12)
        self.external_encoder = QtWidgets.QLineEdit(self.page_external)
        self.external_encoder.setObjectName("external_encoder")
        self.horizontalLayout_4.addWidget(self.external_encoder)
        self.btnChooseEncoder = QtWidgets.QToolButton(self.page_external)
        self.btnChooseEncoder.setObjectName("btnChooseEncoder")
        self.horizontalLayout_4.addWidget(self.btnChooseEncoder)
        self.verticalLayout_6.addLayout(self.horizontalLayout_4)
        spacerItem3 = QtWidgets.QSpacerItem(20, 158, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_6.addItem(spacerItem3)
        self.optionWidget.addWidget(self.page_external)
        self.page_data_whitening = QtWidgets.QWidget()
        self.page_data_whitening.setObjectName("page_data_whitening")
        self.datawhitening_sync = QtWidgets.QLineEdit(self.page_data_whitening)
        self.datawhitening_sync.setGeometry(QtCore.QRect(0, 0, 171, 23))
        self.datawhitening_sync.setObjectName("datawhitening_sync")
        self.label_13 = QtWidgets.QLabel(self.page_data_whitening)
        self.label_13.setGeometry(QtCore.QRect(180, 0, 231, 20))
        self.label_13.setObjectName("label_13")
        self.datawhitening_polynomial = QtWidgets.QLineEdit(self.page_data_whitening)
        self.datawhitening_polynomial.setGeometry(QtCore.QRect(0, 30, 171, 23))
        self.datawhitening_polynomial.setObjectName("datawhitening_polynomial")
        self.label_14 = QtWidgets.QLabel(self.page_data_whitening)
        self.label_14.setGeometry(QtCore.QRect(180, 30, 291, 21))
        self.label_14.setObjectName("label_14")
        self.datawhitening_applycrc = QtWidgets.QCheckBox(self.page_data_whitening)
        self.datawhitening_applycrc.setGeometry(QtCore.QRect(0, 60, 261, 21))
        self.datawhitening_applycrc.setObjectName("datawhitening_applycrc")
        self.datawhitening_preamble_rm = QtWidgets.QCheckBox(self.page_data_whitening)
        self.datawhitening_preamble_rm.setGeometry(QtCore.QRect(0, 80, 221, 21))
        self.datawhitening_preamble_rm.setObjectName("datawhitening_preamble_rm")
        self.datawhitening_sync_rm = QtWidgets.QCheckBox(self.page_data_whitening)
        self.datawhitening_sync_rm.setGeometry(QtCore.QRect(0, 100, 261, 21))
        self.datawhitening_sync_rm.setObjectName("datawhitening_sync_rm")
        self.datawhitening_crc_rm = QtWidgets.QCheckBox(self.page_data_whitening)
        self.datawhitening_crc_rm.setGeometry(QtCore.QRect(0, 120, 131, 21))
        self.datawhitening_crc_rm.setObjectName("datawhitening_crc_rm")
        self.optionWidget.addWidget(self.page_data_whitening)
        self.page_cut = QtWidgets.QWidget()
        self.page_cut.setObjectName("page_cut")
        self.cutmark = QtWidgets.QLineEdit(self.page_cut)
        self.cutmark.setGeometry(QtCore.QRect(0, 30, 181, 31))
        self.cutmark.setObjectName("cutmark")
        self.label_15 = QtWidgets.QLabel(self.page_cut)
        self.label_15.setGeometry(QtCore.QRect(190, 30, 121, 31))
        self.label_15.setObjectName("label_15")
        self.rB_delbefore = QtWidgets.QRadioButton(self.page_cut)
        self.rB_delbefore.setGeometry(QtCore.QRect(0, 0, 131, 23))
        self.rB_delbefore.setObjectName("rB_delbefore")
        self.rB_delafter = QtWidgets.QRadioButton(self.page_cut)
        self.rB_delafter.setGeometry(QtCore.QRect(150, 0, 111, 23))
        self.rB_delafter.setObjectName("rB_delafter")
        self.rB_delbeforepos = QtWidgets.QRadioButton(self.page_cut)
        self.rB_delbeforepos.setGeometry(QtCore.QRect(0, 70, 111, 23))
        self.rB_delbeforepos.setObjectName("rB_delbeforepos")
        self.rB_delafterpos = QtWidgets.QRadioButton(self.page_cut)
        self.rB_delafterpos.setGeometry(QtCore.QRect(150, 70, 111, 23))
        self.rB_delafterpos.setObjectName("rB_delafterpos")
        self.cutmark2 = QtWidgets.QSpinBox(self.page_cut)
        self.cutmark2.setGeometry(QtCore.QRect(0, 100, 181, 33))
        self.cutmark2.setMaximum(1000)
        self.cutmark2.setObjectName("cutmark2")
        self.label_16 = QtWidgets.QLabel(self.page_cut)
        self.label_16.setGeometry(QtCore.QRect(190, 100, 121, 31))
        self.label_16.setObjectName("label_16")
        self.optionWidget.addWidget(self.page_cut)
        self.verticalLayout_5.addWidget(self.optionWidget)
        self.optionWidget.raise_()
        self.info.raise_()
        self.verticalLayout_3.addWidget(self.gb_infoandoptions)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_4.addLayout(self.horizontalLayout)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_3 = QtWidgets.QLabel(Decoder)
        self.label_3.setObjectName("label_3")
        self.gridLayout_2.addWidget(self.label_3, 4, 0, 1, 1)
        self.combobox_signals = QtWidgets.QComboBox(Decoder)
        self.combobox_signals.setObjectName("combobox_signals")
        self.combobox_signals.addItem("")
        self.gridLayout_2.addWidget(self.combobox_signals, 1, 0, 1, 1)
        self.output = QtWidgets.QLineEdit(Decoder)
        self.output.setReadOnly(True)
        self.output.setObjectName("output")
        self.gridLayout_2.addWidget(self.output, 5, 0, 1, 2)
        self.decoding_errors_label = QtWidgets.QLabel(Decoder)
        self.decoding_errors_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.decoding_errors_label.setObjectName("decoding_errors_label")
        self.gridLayout_2.addWidget(self.decoding_errors_label, 4, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(Decoder)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 0, 0, 1, 1)
        self.inpt = QtWidgets.QLineEdit(Decoder)
        self.inpt.setInputMethodHints(QtCore.Qt.ImhDigitsOnly)
        self.inpt.setObjectName("inpt")
        self.gridLayout_2.addWidget(self.inpt, 1, 1, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem4, 0, 1, 1, 1)
        self.graphicsView_signal = ZoomableGraphicView(Decoder)
        self.graphicsView_signal.setObjectName("graphicsView_signal")
        self.gridLayout_2.addWidget(self.graphicsView_signal, 2, 0, 1, 2)
        self.graphicsView_decoded = ZoomableGraphicView(Decoder)
        self.graphicsView_decoded.setObjectName("graphicsView_decoded")
        self.gridLayout_2.addWidget(self.graphicsView_decoded, 3, 0, 1, 2)
        self.verticalLayout_4.addLayout(self.gridLayout_2)

        self.retranslateUi(Decoder)
        self.optionWidget.setCurrentIndex(6)
        QtCore.QMetaObject.connectSlotsByName(Decoder)

    def retranslateUi(self, Decoder):
        _translate = QtCore.QCoreApplication.translate
        Decoder.setWindowTitle(_translate("Decoder", "Decoding"))
        self.combobox_decodings.setItemText(0, _translate("Decoder", "Non Return to Zero (NRZ)"))
        self.combobox_decodings.setItemText(1, _translate("Decoder", "Empty"))
        self.delete_decoding.setText(_translate("Decoder", "Delete"))
        self.saveas.setText(_translate("Decoder", "Save as..."))
        self.label_8.setText(_translate("Decoder", "Base Functions"))
        self.label_9.setText(_translate("Decoder", "Additional Functions"))
        self.label.setText(_translate("Decoder", "Decoder"))
        self.label_4.setText(_translate("Decoder", "Signal"))
        self.label_7.setText(_translate("Decoder", "Decoded Bits"))
        self.gb_infoandoptions.setTitle(_translate("Decoder", "Information and Options"))
        self.info.setText(_translate("Decoder", "Please drag functions from the categories base and additional to the decoding process (Decoder). You can reorder functions by drag and drop and remove functions by dropping them outside the Decoder box. Click on every function for detailed information."))
        self.label_5.setText(_translate("Decoder", "Number of redundant bits"))
        self.label_6.setText(_translate("Decoder", "Carrier (\'1_\' -> 1_1_1_...)"))
        self.label_10.setText(_translate("Decoder", "Rows"))
        self.label_11.setText(_translate("Decoder", "Decoder"))
        self.btnChooseDecoder.setText(_translate("Decoder", "..."))
        self.label_12.setText(_translate("Decoder", "Encoder"))
        self.btnChooseEncoder.setText(_translate("Decoder", "..."))
        self.label_13.setText(_translate("Decoder", "Synchronization bytes (hex coded)"))
        self.label_14.setText(_translate("Decoder", "Data whitening polynomial (LFSR, hex coded)"))
        self.datawhitening_applycrc.setText(_translate("Decoder", "Compute and apply CRC-16 via XOR"))
        self.datawhitening_preamble_rm.setText(_translate("Decoder", "Remove Preamble (1010...)"))
        self.datawhitening_sync_rm.setText(_translate("Decoder", "Remove Synchronization bytes"))
        self.datawhitening_crc_rm.setText(_translate("Decoder", "Remove CRC-16"))
        self.cutmark.setText(_translate("Decoder", "1010"))
        self.label_15.setText(_translate("Decoder", "Sequence"))
        self.rB_delbefore.setText(_translate("Decoder", "&Cut before"))
        self.rB_delafter.setText(_translate("Decoder", "Cut afte&r"))
        self.rB_delbeforepos.setText(_translate("Decoder", "Cut before"))
        self.rB_delafterpos.setText(_translate("Decoder", "Cut after"))
        self.label_16.setText(_translate("Decoder", "Position (in bit)"))
        self.label_3.setText(_translate("Decoder", "Decoded Bits:"))
        self.combobox_signals.setItemText(0, _translate("Decoder", "Test"))
        self.decoding_errors_label.setText(_translate("Decoder", "[Decoding Errors = 0]"))
        self.label_2.setText(_translate("Decoder", "Signal {0,1}:"))

from urh.ui.ListWidget import ListWidget
from urh.ui.views.ZoomableGraphicView import ZoomableGraphicView



from collections import defaultdict
from urh.awre.components.Component import Component
from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.Message import Message


class Preamble(Component):
    """
    Assign Preamble and SoF.

    """
    def __init__(self, fieldtypes, priority=0, predecessors=None, enabled=True, backend=None, messagetypes=None):
        """

        :type fieldtypes: list of FieldType
        :param priority:
        :param predecessors:
        :param enabled:
        :param backend:
        :param messagetypes:
        """
        super().__init__(priority, predecessors, enabled, backend, messagetypes)

        self.preamble_field_type = next((ft for ft in fieldtypes if ft.function == ft.Function.PREAMBLE), None)
        self.sync_field_type = next((ft for ft in fieldtypes if ft.function == ft.Function.SYNC), None)

        self.preamble_name = self.preamble_field_type.caption if self.preamble_field_type else "Preamble"
        self.sync_name = self.sync_field_type.caption if self.sync_field_type else "Synchronization"

    def _py_find_field(self, messages):
        """

        :type messages: list of Message
        :return:
        """
        preamble_ranges = defaultdict(list)
        """:type: dict[MessageType, list] """

        for msg in messages:
            rng = self.__find_preamble_range(msg)
            if rng:
                preamble_ranges[msg.message_type].append(rng)

        preamble_ends = defaultdict(int)
        for message_type, ranges in preamble_ranges.items():
            start, end = max(ranges, key=ranges.count)
            message_type.add_protocol_label(start=start, end=end, name=self.preamble_name,
                                            auto_created=True, type=self.preamble_field_type)

            preamble_ends[message_type] = end + 1

        for message_type in preamble_ranges.keys():
            messages = [msg for msg in messages if msg.message_type == message_type]
            first_field = next((field for field in message_type if field.start > preamble_ends[message_type]), None)
            search_end = first_field.start if first_field is not None else None
            sync_range = self.__find_sync_range(messages, preamble_ends[message_type], search_end)

            if sync_range:
                message_type.add_protocol_label(start=sync_range[0], end=sync_range[1]-1, name=self.sync_name,
                                                auto_created=True, type=self.sync_field_type)


    def __find_preamble_range(self, message: Message):
        search_start = 0

        if len(message.message_type) == 0:
            search_end = len(message.decoded_bits)
        else:
            search_end = message.message_type[0].start

        bits = message.decoded_bits

        # Skip sequences of equal bits
        first_difference = next((i for i in range(search_start, search_end-1) if bits[i] != bits[i+1]), None)
        if first_difference is None:
            return None

        try:
            preamble_end = next((i-1 for i in range(first_difference, search_end, 4)
                if bits[i] == bits[i+1] or bits[i] != bits[i+2] or bits[i] == bits[i+3]), search_end)
        except IndexError:
            return None

        if preamble_end - first_difference > 4:
            return first_difference, preamble_end
        else:
            return None


    def __find_sync_range(self, messages, preamble_end: int, search_end: int):
        """
        Finding the synchronization works by finding the first difference between two messages.
        This is performed for all messages and the most frequent first difference is chosen

        :type messages: list of Message
        :param preamble_end: End of preamble = start of search
        :param search_end: End of search = start of first other label
        """

        possible_sync_pos = defaultdict(int)


        for i, msg in enumerate(messages):
            bits_i = msg.decoded_bits[preamble_end:search_end]
            for j in range(i, len(messages)):
                bits_j = messages[j].decoded_bits[preamble_end:search_end]
                first_diff = next((k for k, (bit_i, bit_j) in enumerate(zip(bits_i, bits_j)) if bit_i != bit_j), None)
                if first_diff is not None:
                    first_diff = preamble_end + 4 * (first_diff // 4)
                    if (first_diff - preamble_end) >= 4:
                        possible_sync_pos[(preamble_end, first_diff)] += 1
        try:
            sync_interval = max(possible_sync_pos, key=possible_sync_pos.__getitem__)
            return sync_interval
        except ValueError:
            return None


from abc import ABCMeta

from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from enum import Enum

from urh.util.Logger import logger


class Component(metaclass=ABCMeta):
    """
    A component is the basic building block of our AWRE algorithm.
    A component can be a Preamble or Sync or Length Field finding routine.
    Components can have a priority which determines the order in which they are processed by the algorithm.
    Additionally, components can have a set of predecessors to define hard dependencies.
    """


    EQUAL_BIT_TRESHOLD = 0.9

    class Backend(Enum):
        python = 1
        cython = 2
        plainc = 3

    def __init__(self, priority=0, predecessors=None, enabled=True, backend=None, messagetypes=None):
        """

        :param priority: Priority for this Component. 0 is highest priority
        :type priority: int
        :param predecessors: List of preceding components, that need to be run before this one
        :type predecessors: list of Component or None
        :param messagetypes: Message types of the examined protocol
        :type messagetypes: list[MessageType]
        """
        self.enabled = enabled
        self.backend = backend if backend is not None else self.Backend.python
        self.priority = abs(priority)
        self.predecessors = predecessors if isinstance(predecessors, list) else []
        """:type: list of Component """

        self.messagetypes = messagetypes

    def find_field(self, messages):
        """
        Wrapper method selecting the backend to assign the protocol field.
        Various strategies are possible e.g.:
        1) Heuristics e.g. for Preamble
        2) Scoring based e.g. for Length
        3) Fulltext search for addresses based on participant subgroups

        :param messages: messages a field shall be searched for
        :type messages: list of Message
        """
        try:
            if self.backend == self.Backend.python:
                self._py_find_field(messages)
            elif self.backend == self.Backend.cython:
                self._cy_find_field(messages)
            elif self.backend == self.Backend.plainc:
                self._c_find_field(messages)
            else:
                raise ValueError("Unsupported backend {}".format(self.backend))
        except NotImplementedError:
            logger.info("Skipped {} because not implemented yet".format(self.__class__.__name__))

    def _py_find_field(self, messages):
        raise NotImplementedError()

    def _cy_find_field(self, messages):
        raise NotImplementedError()

    def _c_find_field(self, messages):
        raise NotImplementedError()


    def assign_messagetypes(self, messages, clusters):
        """
        Assign message types based on the clusters. Following rules:
        1) Messages from different clusters will get different message types
        2) Messages from same clusters will get same message type
        3) The new message type will copy over the existing labels
        4) No new message type will be set for messages, that already have a custom message type assigned

        For messages with clustername "default" no new message type will be created

        :param messages: Messages, that messagetype needs to be clustered
        :param clusters: clusters for the messages
        :type messages: list[Message]
        :type clusters: dict[str, set[int]]
        :return:
        """
        for clustername, clustercontent in clusters.items():
            if clustername == "default":
                # Do not force the default message type
                continue

            for msg_i in clustercontent:
                msg = messages[msg_i]
                if msg.message_type == self.messagetypes[0]:
                    # Message has default message type
                    # Copy the existing labels and create a new message type
                    # if it was not already done
                    try:
                        msg_type = next(mtype for mtype in self.messagetypes if mtype.name == clustername)
                    except StopIteration:
                        msg_type = MessageType(name=clustername, iterable=msg.message_type)
                        msg_type.assigned_by_logic_analyzer = True
                        self.messagetypes.append(msg_type)
                    msg.message_type = msg_type




from collections import defaultdict

import numpy
from PyQt5.QtCore import QAbstractTableModel, QModelIndex, Qt
from PyQt5.QtGui import QFont, QColor
from PyQt5.QtWidgets import QUndoStack, QMessageBox

from urh import constants
from urh.cythonext.signalFunctions import Symbol


class TableModel(QAbstractTableModel):
    def __init__(self, participants, parent=None):
        super().__init__(parent)
        self.controller = None
        """:type: CompareFrameController or GeneratorTabController """
        self.protocol = None # Reimplement in Child classes
        """:type: ProtocolAnalyzer """

        self.col_count = 0
        self.row_count = 0
        self.display_data = None
        """:type: list[str] """

        self.search_results = []
        self.search_value = ""
        self._proto_view = 0
        self._refindex = -1

        self.first_messages = []
        self.hidden_rows = set()

        self.is_writeable = False
        self.locked = False
        self.decode = True  # False for Generator

        self.symbols = {}
        """:type: dict[str, Symbol] """

        self.background_colors = defaultdict(lambda: None)
        self.bold_fonts = defaultdict(lambda: False)
        self.text_colors = defaultdict(lambda: None)
        self.tooltips = defaultdict(lambda: None)
        self.vertical_header_text = defaultdict(lambda: None)
        self.vertical_header_colors = defaultdict(lambda: None)

        self._diffs = defaultdict(set)
        """:type: dict[int, set[int]] """

        self.undo_stack = QUndoStack()

        self.__participants = participants

    @property
    def participants(self):
        return self.__participants

    @participants.setter
    def participants(self, value):
        self.__participants = value
        for msg in self.protocol.messages:
            if msg.participant not in self.__participants:
                msg.participant = None


    @property
    def proto_view(self):
        return self._proto_view

    @proto_view.setter
    def proto_view(self, value):
        self._proto_view = value
        if self._refindex >= 0:
            self._diffs = self.protocol.find_differences(self._refindex, self._proto_view)
        self.update()

    def headerData(self, section: int, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Vertical:
            if role == Qt.DisplayRole:
                return self.vertical_header_text[section]
            elif role == Qt.BackgroundColorRole:
                return self.vertical_header_colors[section]
            elif role == Qt.TextColorRole:
                color = self.vertical_header_colors[section]
                if color:
                    red, green, blue  = color.red(), color.green(), color.blue()
                    return QColor("black") if (red * 0.299 + green * 0.587 + blue * 0.114) > 186 else QColor("white")
                else:
                    return None

        return super().headerData(section, orientation, role)

    def update(self):
        self.locked = True

        self.symbols.clear()
        self.symbols = {symbol.name: symbol for symbol in self.protocol.used_symbols}

        if self.protocol.num_messages > 0:
            if self.decode:
                if self.proto_view == 0:
                    self.display_data = self.protocol.decoded_proto_bits_str
                elif self.proto_view == 1:
                    self.display_data = self.protocol.decoded_hex_str
                elif self.proto_view == 2:
                    self.display_data = self.protocol.decoded_ascii_str
            else:
                #
                # Generator Model
                if self.proto_view == 0:
                    self.display_data = self.protocol.plain_bits_str
                elif self.proto_view == 1:
                    self.display_data = self.protocol.plain_hex_str
                else:
                    self.display_data = self.protocol.plain_ascii_str

            visible_messages = [msg for i, msg in enumerate(self.display_data) if i not in self.hidden_rows]
            if len(visible_messages) == 0:
                self.col_count = 0
            else:
                self.col_count = numpy.max([len(msg) for msg in visible_messages])

            if self._refindex >= 0:
                self._diffs = self.protocol.find_differences(self._refindex, self.proto_view)
            else:
                self._diffs.clear()

            self.row_count = self.protocol.num_messages
            self.find_protocol_value(self.search_value)
        else:
            self.col_count = 0
            self.row_count = 0
            self.display_data = None

        # Cache background colors for performance
        self.refresh_bgcolors_and_tooltips()
        self.refresh_fonts()  # Will be overriden
        self.refresh_vertical_header()

        self.beginResetModel()
        self.endResetModel()
        self.locked = False

    def columnCount(self, QModelIndex_parent=None, *args, **kwargs):
        return self.col_count

    def rowCount(self, QModelIndex_parent=None, *args, **kwargs):
        return self.row_count

    def refresh_bgcolors_and_tooltips(self):
        self.background_colors.clear()
        self.tooltips.clear()
        label_colors = constants.LABEL_COLORS

        for i, message in enumerate(self.protocol.messages):
            for lbl in message.message_type:
                bg_color = label_colors[lbl.color_index]
                start, end = message.get_label_range(lbl, self.proto_view, self.decode)
                for j in range(start, end):
                    self.background_colors[i, j] = bg_color
                    self.tooltips[i, j] = lbl.name

    def refresh_fonts(self):
        """
        Will be overriden

        :return:
        """
        pass

    def refresh_vertical_header(self):
        self.vertical_header_colors.clear()
        self.vertical_header_text.clear()
        for i in range(self.row_count):
            try:
                participant = self.protocol.messages[i].participant
            except IndexError:
                participant = None
            if participant:
                self.vertical_header_text[i] = "{0} ({1})".format(i + 1, participant.shortname)
                self.vertical_header_colors[i] = constants.PARTICIPANT_COLORS[participant.color_index]
            else:
                self.vertical_header_text[i] = str(i+1)


    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        i = index.row()
        j = index.column()
        if role == Qt.DisplayRole and self.display_data:
            try:
                return self.display_data[i][j]
            except IndexError:
                return None

        elif role == Qt.TextAlignmentRole:
            if i in self.first_messages:
                return Qt.AlignHCenter + Qt.AlignBottom
            else:
                return Qt.AlignCenter

        elif role == Qt.BackgroundColorRole:
            return self.background_colors[i, j]

        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(self.bold_fonts[i, j])
            return font

        elif role == Qt.TextColorRole:
            return self.text_colors[i, j]

        elif role == Qt.ToolTipRole:
            return self.tooltips[i, j]

        else:
            return None

    def setData(self, index: QModelIndex, value, role=Qt.DisplayRole):
        if role == Qt.EditRole:
            i = index.row()
            j = index.column()
            hex_chars = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "a", "b", "c", "d", "e", "f")

            if self.proto_view == 0:
                if value in ("0", "1"):
                    try:
                        self.protocol.messages[i][j] = bool(int(value))
                        self.update()
                    except IndexError:
                        return False
                elif value in self.symbols.keys():
                    try:
                        self.protocol.messages[i][j] = self.symbols[value]
                        self.update()
                    except IndexError:
                        return False
            elif self.proto_view == 1:
                if value in hex_chars:
                    index = self.protocol.convert_index(j, 1, 0, True, message_indx=i)[0]
                    bits = "{0:04b}".format(int(value, 16))
                    for k in range(4):
                        try:
                            if type(self.protocol.messages[i][index + k]) != Symbol:
                                self.protocol.messages[i][index + k] = bool(int(bits[k]))
                            else:
                                break
                        except IndexError:
                            break
                    self.update()
                elif value in self.symbols.keys():
                    QMessageBox.information(None, "Setting symbol", "You can only set custom bit symbols in bit view!")
            elif self.proto_view == 2 and len(value) == 1:
                if value in self.symbols.keys():
                    QMessageBox.information(None, "Setting symbol", "You can only set custom bit symbols in bit view!")

                index = self.protocol.convert_index(j, 2, 0, True, message_indx=i)[0]
                bits = "{0:08b}".format(ord(value))
                for k in range(8):
                    try:
                        if type(self.protocol.messages[i][index + k]) != Symbol:
                            self.protocol.messages[i][index + k] = bool(int(bits[k]))
                        else:
                            break
                    except IndexError:
                        break
                self.update()
        return True

    def find_protocol_value(self, value):
        self.search_results[:] = []
        self.search_value = value

        if len(value) == 0:
            return 0

        for i, message in enumerate(self.display_data):
            if i in self.hidden_rows:
                continue

            j = message.find(value)
            while j != -1:
                self.search_results.append((i, j))
                j = message.find(value, j + 1)

        return len(self.search_results)